""" Fixtures and functions for generating synthetic data for testing """
import pytest
import numpy as np

from scipy.stats import norm, poisson, bernoulli
from scipy.stats.mstats import gmean

from spynal.spectra.utils import set_random_seed, simulate_oscillation

# Possible errors to expect when inputting a missing/misspelled argument
# Used for unit tests against silently ignoring incorrect arguments
MISSING_ARG_ERRS = (TypeError,AttributeError,AssertionError)


# =============================================================================
# Fixtures for generating random test data of different data schemes
# =============================================================================
@pytest.fixture(scope='session')
def one_sample_data():
    """
    Fixture simulates one-sample (single-condition) random data to use for unit tests

    RETURNS
    data    (10,4) ndarray. Simulated data, simulating (10 trials x 4 channels)
    """
    # Note: seed=1 makes data reproducibly match output of Matlab
    set_random_seed(1)

    n       = 10
    n_chnls = 4
    mu      = 10.0
    sd      = 5.0
    return norm.ppf(np.random.rand(n,n_chnls), loc=mu, scale=sd)


@pytest.fixture(scope='session')
def two_sample_data():
    """
    Fixture simulates set of two-sample (2 condition) random data to use for unit tests

    RETURNS
    data    (20,4) ndarray. Simulated data, simulating (10 trials*2 conditions/levels x 4 channels)

    labels  (20,) ndarray of int8. Set of condition labels corresponding to each trial in data
    """
    # Note: seed=1 makes data reproducibly match output of Matlab
    set_random_seed(1)

    n       = 10
    n_chnls = 4
    mu      = (10.0, 20.0)
    sd      = 5.0
    data = np.concatenate((norm.ppf(np.random.rand(n,n_chnls), loc=mu[0], scale=sd),
                           norm.ppf(np.random.rand(n,n_chnls), loc=mu[1], scale=sd)), axis=0)
    labels = np.hstack((np.zeros((n,),dtype='uint8'), np.ones((n,),dtype='uint8')))

    return data, labels


@pytest.fixture(scope='session')
def one_way_data():
    """
    Fixture simulates set of 3-condition random data to use for unit tests of one-way test functions

    RETURNS
    data    (30,4) ndarray. Simulated data, simulating (10 trials*3 conditions x 4 channels)

    labels  (30,) ndarray of int8. Set of condition labels corresponding to each trial in data
    """
    # Note: seed=1 makes data reproducibly match output of Matlab
    set_random_seed(1)

    n       = 10
    n_groups= 3
    n_chnls = 4
    mu      = (10.0, 20.0, 30.0)
    sd      = 5.0
    data = np.concatenate([norm.ppf(np.random.rand(n,n_chnls), loc=mu[i], scale=sd)
                           for i in range(n_groups)], axis=0)
    labels = np.hstack([i*np.ones((n,),dtype='uint8') for i in range(3)])

    return data, labels


@pytest.fixture(scope='session')
def two_way_data():
    """
    Fixture simulates set of 4-condition random data to use for unit tests of two-way test functions

    RETURNS
    data    (40,4) ndarray. Simulated data, simulating (10 trials*4 conditions/levels x 4 channels)

    labels  (40,4) ndarray of int8. Set of condition labels corresponding to each trial in data.
            Columns 0,1 correspond to main effects, column 2 corresponds to interaction
    """
    # Note: seed=1 makes data reproducibly match output of Matlab
    set_random_seed(1)

    n       = 10
    n_groups= 4
    n_terms = 3
    n_chnls = 4
    mu      = (10.0, 20.0, 30.0, 40.0)
    sd      = 5.0
    data = np.concatenate([norm.ppf(np.random.rand(n,n_chnls), loc=mu[i], scale=sd)
                           for i in range(n_groups)], axis=0)
    labels = np.empty((n*n_groups,n_terms), dtype='uint8')
    labels[:,0] = np.tile(np.hstack((np.zeros((n,)), np.ones((n,)))), (2,))
    labels[:,1] = np.hstack((np.zeros((n*2,)), np.ones((n*2,))))
    labels[:,2] = np.hstack([i*np.ones((n,)) for i in range(n_groups)])

    return data, labels


# =============================================================================
# Fixtures for generating oscillatory test data
# =============================================================================
@pytest.fixture(scope='session')
def oscillation():
    """
    Fixture simulates set of instances of oscillatory data for all unit tests

    RETURNS
    data    (1000,4) ndarray. Simulated oscillatory data.
            (eg simulating 1000 timepoints x 4 trials or channels)
    """
    # Note: seed=1 makes data reproducibly match output of Matlab
    frequency = 32
    return simulate_oscillation(frequency, amplitude=5.0, phase=0, noise=1.0,
                                n_trials=4, time_range=1.0, smp_rate=1000, seed=1)


@pytest.fixture(scope='session')
def bursty_oscillation():
    """
    Fixture simulates set of instances of bursty oscillatory data for all unit tests

    RETURNS
    data    (1000,4) ndarray. Simulated bursty oscillatory data.
            (eg simulating 1000 timepoints x 4 trials or channels)
    """
    # Note: seed=1 makes data reproducibly match output of Matlab
    frequency = 32
    return simulate_oscillation(frequency, amplitude=5.0, phase=0, noise=1.0, burst_rate=0.4,
                                time_range=1.0, n_trials=4, smp_rate=1000, seed=1)


@pytest.fixture(scope='session')
def spiking_oscillation(oscillation):
    """
    Fixture simulates set of instances of oscillatory spiking data for all unit tests

    RETURNS
    data    (1000,4) ndarray of bool. Simulated oscillatory spiking data,
            expressed as binary (0/1) spike trains.
            (eg simulating 1000 timepoints x 4 trials or channels)
    """
    # todo code up something actually proper (rate-modulated Poisson process?)
    data = oscillation

    # Convert continuous oscillation to probability (range 0-1)
    data = (data - data.min()) / data.ptp()
    data = data**2  # Sparsen high rates some

    # Use probabilities to generate Bernoulli random variable at each time point
    return bernoulli.ppf(0.5, data).astype(bool)


@pytest.fixture(scope='session')
def oscillatory_data(oscillation, bursty_oscillation, spiking_oscillation):
    """
    "Meta-fixture" that returns standard analog oscillations, bursty oscillations,
    and spiking oscillations in a single dict.

    RETURNS
    data_dict   {'data_type' : data} dict containing outputs from
                each of constituent fixtures

    SOURCE      https://stackoverflow.com/a/42400786
    """
    return {'lfp': oscillation, 'burst': bursty_oscillation, 'spike':spiking_oscillation}


# =============================================================================
# Functions for generating simulated data
# =============================================================================
def simulate_data(distribution='normal', mean=None, spread=1, n=100, seed=None):
    """
    Simulates random data with given distribution and parameters

    data = simulate_data(distribution='normal', mean=None, spread=1, n=100, seed=None)

    ARGS
    distribution    String. Name of distribution to simulate data from.
                    Options: 'normal' [default] | 'poisson'

    mean            Scalar. Mean/rate parameter for distribution.
                    Default: 0 if distribution='normal' else 10

    spread          Scalar. Spread parameter for distribution.
                    SD for normal; unused for Poisson
                    Default: 1.0

    n               Int | Tuple. Number/shape of random variable draws to simulate.
                    Can set either single integer value or a tuple for shape of RV array.
                    Default: 100

    seed            Int. Random generator seed for repeatable results.
                    Set=None [default] for unseeded random numbers.

    RETURNS
    data            (n,) | (*n) ndarray. Simulated random data.
                    Returns as 1D array if n is an int.
                    Returns with shape given by n if it is a tuple.
    """
    if seed is not None: set_random_seed(seed)

    if mean is None:
        if distribution in ['normal','norm','gaussian','gauss']:    mean = 0.0
        else:                                                       mean = 10

    dist_func = _distribution_name_to_func(distribution)

    return dist_func(n, mean, spread)


def simulate_dataset(gain=5.0, offset=5.0, n_conds=2, n=100, n_chnls=1, distribution='normal',
                     spreads=1.0, correlation=0, seed=None):
    """
    Simulates random data across multiple conditions/groups with given condition effect size,
    distribution and parameters

    data,labels = simulate_dataset(gain=5.0,offset=5.0,n_conds=2,n=100,
                                   distribution='normal',correlation=0,seed=None)

    ARGS
    gain    Scalar | (n_conds,) array-like. Sets the effect size (difference
            in mean value between each condition).
            If scalar, it's the difference in mean between each successive condition.
            If vector, it's the difference in mean from baseline for each individual condition.
            Set = 0 to simulate no expected difference between conditions.
            Default: 5.0

    offset  Scalar. Additive baseline value added to any condition effects. Default: 5.0

    n_conds Int. Number of distinct conditions/groups to simulate. Default: 2

    n       Int. Number of trials/observations to simulate *per condition*. Default: 100
    
    n_chnls Int. Number of independent data channels to simulate (all channels currently
            have same stochastic properties). Default: 1

    distribution    String. Name of distribution to simulation from.
                    Options: 'normal' [default] | 'poisson'

    spreads Scalar | (n_conds,) array-like. Spread of each condition's data distribution.
            (Gaussian SD; not used for other distributions).
            If scalar, the same spread is used for all conditions.
            If vector, one spread value should be given for each condition.

    correlation Float in range[-1,+1]. Correlation between data in each condition.
            Note: Currently only supported for 2 conditions with normal distribution
            (simulated as multivariate normal w/ covariance matrix based on correlation,spreads)

    seed    Int. Random generator seed for repeatable results.
            Set=None [default] for unseeded random numbers.

    RETURNS
    data    (n*n_conds,n_chnls). Simulated data for multiple repetitions of one/more conditions.

    labels  (n*n_conds,) of int. Condition/group labels for each trial.
            Sorted in group order to simplify visualization.
    """
    # todo Add ability to simulate independent data series, different n for each cond,
    if seed is not None: set_random_seed(seed)

    # For single-condition data, treat gain as scalar increase over baseline response
    if n_conds == 1:
        gains = np.asarray([gain])
        spreads = np.asarray([spreads])

    else:
        # Single gain = incremental difference btwn cond 0 and 1, 1 and 2, etc.
        # Convert to vector of gains for each condition
        gains = gain*np.arange(n_conds) if np.isscalar(gain) else gain

        # Convert scalar spread value to vector for each condition
        if np.isscalar(spreads): spreads = spreads*np.ones((n_conds,))

    assert len(gains) == n_conds, \
        ValueError("Vector-valued <gain> must have length == n_conds (%d != %d)" \
                    % (len(gains), n_conds))

    assert len(spreads) == n_conds, \
        ValueError("Vector-valued <spreads> must have length == n_conds (%d != %d)" \
                    % (len(spreads), n_conds))

    assert (correlation >= -1) and (correlation <= 1), \
        ValueError("Correlation must be in range [-1,+1] (%.2f input)" % correlation)

    if correlation != 0:
        assert (n_conds == 2) and (distribution == 'normal'), \
            ValueError("correlation currently only supported for 2 conds, normal distribution")

    # Final mean value = baseline + condition-specific gain
    means = offset + gains

    # Generate data for each condition and stack together -> (n_trials*n_conds,) array
    if correlation == 0:
        n_ = n if n_chnls == 1 else (n,n_chnls)
        data = np.concatenate([simulate_data(distribution=distribution, mean=mean, spread=spread, n=n_, seed=None)
                               for mean,spread in zip(means,spreads)], axis=0)

    # Generate data for both condition using single multivariate normal distribution
    else:
        # Convert SDs -> variances and compute pooled variance = geometric mean
        variances = np.asarray(spreads)**2
        var_pooled = gmean(variances)
        # Covariance matrix = variances and covariance = pooled variance * correlation
        cov_mx = [[variances[0], var_pooled*correlation], [var_pooled*correlation, variances[1]]]

        # Generate multivariate normal data with given means and covariance matrix
        n_ = (n,) if n_chnls == 1 else (n,n_chnls)
        data = np.random.multivariate_normal(means, cov_mx, size=n_)
        # Reshape (n,n_conds) -> (n*n_conds,)
        data = data.reshape((n*n_conds,), order='F')

    # Create (n_trials_total,) vector of group labels (ints in set 0-n_conds-1)
    labels = np.hstack([i_cond*np.ones((n,)) for i_cond in range(n_conds)])

    return data, labels


# =============================================================================
# Helper functions
# =============================================================================
def _distribution_name_to_func(name):
    """ Converts name of distribution to scipy.stats-based function to generate random variables """
    # Random variables generated in a way that reproducibly matches output of Matlab
    name = name.lower()

    # Normal RV's : mu = mean, s = SD
    if name in ['normal','norm','gaussian','gauss']:
        return lambda n,mu,s: norm.ppf(np.random.rand(n) if np.isscalar(n) else np.random.rand(*n),
                                       loc=mu, scale=s)

    # Poisson RV's : mu = lamba (aka mean,rate), s is unused
    elif name in ['poisson','poiss']:
        return lambda n,mu,s: poisson.ppf(np.random.rand(n) if np.isscalar(n) else np.random.rand(*n),
                                          mu=mu)

    else:
        raise ValueError("%s distribution is not yet supported. Should be 'normal' | 'poisson'")



# -*- coding: utf-8 -*-
""" Multitaper spectral analysis """
from math import floor, sqrt
import numpy as np

from scipy.signal.windows import dpss

from spynal.utils import iarange
from spynal.spikes import _spike_data_type, times_to_bool
from spynal.spectra.preprocess import remove_dc
from spynal.spectra.utils import next_power_of_2, get_freq_sampling, complex_to_spec_type, phase
from spynal.spectra.helpers import fft, _extract_triggered_data


def multitaper_spectrum(data, smp_rate, axis=0, data_type='lfp', spec_type='complex',
                        freq_range=None, removeDC=True, freq_width=4, n_tapers=None,
                        keep_tapers=False, tapers=None, pad=True, **kwargs):
    """
    Multitaper Fourier spectrum computation for continuous (eg LFP) or point process (spike) data

    Multitaper methods project the data onto orthogonal Slepian (DPSS) "taper" functions, which
    increases the data's effective signal-to-noise. It allows a principled tradeoff btwn time
    resolution (data.shape[axis]), frequency resolution (`freq_width`), and the number of taper
    functions (`n_tapers`), which determines the signal-to-noise boost.

    Note: By default, data is zero-padded to the next power of 2 greater than its input length.
    This will change the frequency sampling (number of freqs and exact freqs sampled) from what
    would be obtained from the original raw data, but can be skipped by inputtng pad=False.

    Only parameters differing from :func:`.spectrum` are described here.

    Parameters
    ----------
    freq_range : array-like, shape=(2,) or scalar, default: all frequencies from FFT (0-smp_rate/2)
        Range of frequencies to keep in output, either given as an explicit [low,high]
        range or just a scalar giving the highest frequency to return.

    freq_width : scalar, default: 4 Hz
        Frequency bandwidth 'W' (Hz).

    n_tapers : scalar, default: (2TW-1)
        Number of tapers to compute. Must be <= 2TW-1, as this is the max number of
        spectrally delimited tapers (and is set as default based on set T,W values).
        Note: Time bandwidth 'T' is set to full length of data.

    tapers : ndarray, shape=(n_win_samples,n_tapers), default: (computed from t/f_range,ntapers)
        Precomputed tapers (as computed by :func:`compute_tapers`).

        Alternative method for explictly setting taper functions.
        Input either `time_width`/`freq_width`/`n_tapers` OR `tapers`.
        If tapers not explicitly input, we compute them from `time_width`/`freq_width`/`n_tapers`.
        If tapers *are* explicitly input, `time_width`/`freq_width`/`n_tapers` are ignored.

    keep_tapers : bool, default: False
        If True, tapers axis is retained in output, between frequency axis and time `axis`.
        If False, output is averaged over tapers, and taper axis is removed

    pad : bool, default: True
        If True, zero-pads data to next power of 2 length

    Returns
    -------
    spec : ndarray, shape=(...,n_freqs,[n_tapers,]...), dtype=complex or float
        Multitaper spectrum of given type of data. Sampling (time) axis is
        replaced by frequency and taper axes (if `keep_tapers` is True), but
        shape is otherwise preserved.
        dtype is complex if `spec_type` is 'complex', float otherwise.

    freqs : ndarray, dtype=(n_freqs,)
        List of frequencies in `spec` (in Hz)

    References
    ----------
    - Mitra & Pesaran 1999 https://doi.org/10.1016/S0006-3495(99)77236-X
    - Jarvis & Mitra 2001 https://doi.org/10.1162/089976601300014312
    """
    if axis < 0: axis = data.ndim + axis

    # Convert spike timestamp data to boolean spike train format
    if (data_type == 'spike') and (_spike_data_type(data) == 'timestamp'):
        lims    = kwargs.pop('lims',None)
        bins    = kwargs.pop('bins',None)
        data,_  = times_to_bool(data, width=1/smp_rate, lims=lims, bins=bins)
        axis    = data.ndim
    assert len(kwargs) == 0, \
        TypeError("Incorrect or misspelled variable(s) in keyword args: "+', '.join(kwargs.keys()))

    # If observation axis != 0, permute axis to make it so
    if axis != 0: data = np.moveaxis(data,axis,0)

    n_timepts = data.shape[0]
    # Set FFT length = data length if no padding; else pad to next power of two
    if not pad: n_fft = n_timepts
    else:       n_fft = next_power_of_2(n_timepts)
    # Set frequency sampling vector
    freqs,fbool = get_freq_sampling(smp_rate,n_fft,freq_range=freq_range)

    # Compute DPSS taper functions (if not precomputed)
    if tapers is None:
        tapers = compute_tapers(smp_rate, time_width=n_timepts/smp_rate,
                                freq_width=freq_width, n_tapers=n_tapers)

    # Reshape tapers to (n_timepts,n_tapers) (if not already)
    if (tapers.ndim == 2) and (tapers.shape[1] == n_timepts): tapers = tapers.T
    assert tapers.shape[0] == n_timepts, \
        ValueError("tapers must have same length (%d) as number of timepoints in data (%d)"
                   % (tapers.shape[0],n_timepts))

    # Reshape tapers array to pad end of it w/ singleton dims
    taper_shape  = (*tapers.shape,*np.ones((data.ndim-1,),dtype=int))

    # DELETE Results are identical with just subtracting of DC from data before fft
    # # Compute values needed for normalizing point process (spiking) signals
    # if data_type == 'spike' and removeDC:
    #     # Compute Fourier transform of tapers
    #     taper_fft= fft(tapers,n=n_fft,axis=0)
    #     if data.ndim > 1:
    #         taper_fft_shape = list(taper_shape)
    #         taper_fft_shape[0] = n_fft
    #         taper_fft = np.reshape(taper_fft,taper_fft_shape)
    #     # Compute mean spike rate across all timepoints in each data series
    #     mean_rate = np.sum(data,axis=0,keepdims=True)/n_timepts

    # Reshape tapers and data to have appropriate shapes to broadcast together
    if data.ndim > 1:  tapers = np.reshape(tapers,taper_shape)

    if removeDC: data = remove_dc(data,axis=0)

    # Insert dimension for tapers in data axis 1 -> (n_timepts,1,...)
    data    = data[:,np.newaxis,...]

    # Project data onto set of taper functions
    data    = data * tapers

    # Compute Fourier transform of projected data, normalizing appropriately
    spec    = fft(data,n=n_fft,axis=0)
    if data_type != 'spike': spec = spec/smp_rate

    # DELETE Results are identical with just subtracting of DC from data before fft
    # Subtract off the DC component (average spike rate) for point process signals
    # if data_type == 'spike' and removeDC: spec -= taper_fft*mean_rate

    # Extract desired set of frequencies
    spec    = spec[fbool,...]

    # Convert to desired output spectral signal type
    spec    = complex_to_spec_type(spec,spec_type)

    # Compute mean across tapers if requested
    if not keep_tapers:
        if spec_type == 'phase':    spec = phase(np.exp(1j*spec).mean(axis=1)) # circular mean
        else:                       spec = spec.mean(axis=1)

    # If observation axis wasn't 0, permute (freq,tapers) back to original position
    if axis != 0:
        if keep_tapers: spec = np.moveaxis(spec,[0,1],[axis,axis+1])
        else:           spec = np.moveaxis(spec,0,axis)

    return spec, freqs


def multitaper_spectrogram(data, smp_rate, axis=0, data_type='lfp', spec_type='complex',
                           freq_range=None, removeDC=True, time_width=0.5, freq_width=4,
                           n_tapers=None, spacing=None, tapers=None, keep_tapers=False,
                           pad=True, **kwargs):
    """
    Compute multitaper time-frequency spectrogram for continuous (eg LFP)
    or point process (eg spike) data

    Multitaper methods project the data onto orthogonal Slepian (DPSS) "taper" functions, which
    increases the data's effective signal-to-noise. It allows a principled tradeoff btwn time
    resolution (data.shape[axis]), frequency resolution (freq_width), and the number of taper
    functions (n_tapers), which determines the signal-to-noise increase.

    Note: By default, data is zero-padded to the next power of 2 greater than its input length.
    This will change the frequency sampling (number of freqs and exact freqs sampled) from what
    would be obtained from the original raw data, but can be skipped by inputtng pad=False.

    Only parameters differing from :func:`.spectrogram` are described here.

    Parameters
    ----------
    freq_range : array-like, shape=(2,) or scalar, default: all frequencies from FFT (0-smp_rate/2)
        Range of frequencies to keep in output, either given as an explicit [low,high]
        range or just a scalar giving the highest frequency to return.

    time_width : scalar, default: 0.5 (500 ms)
        Time bandwidth 'T' (s). Width of sliding time window is set equal to this.

    freq_width : scalar, default: 4 Hz
        Frequency bandwidth 'W' (Hz).

    n_tapers : scalar, default: (2TW-1)
        Number of tapers to compute. Must be <= 2TW-1, as this is the max number of
        spectrally delimited tapers (and is set as default based on set T,W values).

    spacing : scalar, default: `time_width` (so each window exactly non-overlapping)
        Spacing between successive sliding time windows (s)

    tapers : ndarray, shape=(n_win_samples,n_tapers), default: (computed from t/f_range,ntapers)
        Precomputed tapers (as computed by :func:`compute_tapers`).

        Alternative method for explicitly setting taper functions.
        Input either `time_width`/`freq_width`/`n_tapers` OR `tapers`.
        If tapers not explicitly input, we compute them from `time_width`/`freq_width`/`n_tapers`.
        If tapers *are* explicitly input, `time_width`/`freq_width`/`n_tapers` are ignored.

    keep_tapers : bool, default: False
        If True, tapers axis is retained in output, between frequency axis and time `axis`.
        If False, output is averaged over tapers, and taper axis is removed

    pad : bool, default: True
        If True, zero-pads data to next power of 2 length

    Returns
    -------
    spec : ndarray, shape=(...,n_freqs[,n_tapers],n_timewins,...), dtype=complex or float
        Multitaper time-frequency spectrogram of data.
        Sampling (time) axis is replaced by frequency, taper (if keep_tapers=True),
        and time window axes but shape is otherwise preserved.
        dtype is complex if `spec_type` is 'complex', float otherwise.

    freqs : ndarray, shape=(n_freqs,)
        List of frequencies in `spec` (in Hz)

    timepts : ndarray, shape=(n_timewins,...)
        List of timepoints in `spec` (in s, referenced to start of data).
        Timepoints here are centers of each time window.

    References
    ----------
    - Mitra & Pesaran 1999 https://doi.org/10.1016/S0006-3495(99)77236-X
    - Jarvis & Mitra 2001 https://doi.org/10.1162/089976601300014312
    """
    if axis < 0: axis = data.ndim + axis

    # Convert spike timestamp data to boolean spike train format
    if (data_type == 'spike') and (_spike_data_type(data) == 'timestamp'):
        lims    = kwargs.pop('lims',None)
        bins    = kwargs.pop('bins',None)
        data,_  = times_to_bool(data, width=1/smp_rate, lims=lims, bins=bins)
        axis    = data.ndim

    # If observation axis != 0, permute axis to make it so
    if axis != 0: data = np.moveaxis(data,axis,0)
    n_timepts = data.shape[0]

    window = time_width
    if spacing is None: spacing = window
    # Compute DPSS taper functions (if not precomputed)
    if tapers is None:
        tapers = compute_tapers(smp_rate, time_width=time_width, freq_width=freq_width,
                                n_tapers=n_tapers)

    # Set up parameters for data time windows
    # Set window starts to range from time 0 to time n - window width
    win_starts  = iarange(0, n_timepts/smp_rate - window, spacing)
    # Set sampled timepoint vector = center of each window
    timepts     = win_starts + window/2.0

    # Extract time-windowed version of data -> (n_timepts_per_win,n_wins,n_dataseries)
    data = _extract_triggered_data(data, smp_rate, win_starts, [0,window])

    if removeDC: data = remove_dc(data, axis=0)

    # Do multitaper analysis on windowed data
    # Note: Set axis=0 and removeDC=False bc already dealt with above.
    # Note: Input values for `freq_width`,`n_tapers` are implicitly propagated here via `tapers`
    spec, freqs = multitaper_spectrum(data, smp_rate, axis=0, data_type=data_type,
                                      spec_type=spec_type, freq_range=freq_range, tapers=tapers,
                                      pad=pad, removeDC=False, keep_tapers=keep_tapers, **kwargs)

    # If time axis wasn't 0, permute (freq,tapers,timewin) axes back to original position
    if axis != 0:
        if keep_tapers: spec = np.moveaxis(spec, [0,1,2], [axis,axis+1,axis+2])
        else:           spec = np.moveaxis(spec, [0,1], [axis,axis+1])

    return spec, freqs, timepts


def compute_tapers(smp_rate, time_width=0.5, freq_width=4, n_tapers=None):
    """
    Compute Discrete Prolate Spheroidal Sequence (DPSS) tapers for use in
    multitaper spectral analysis.

    Uses scipy.signal.windows.dpss, but arguments are different here

    Parameters
    ----------
    smp_rate : scalar
        Data sampling rate (Hz)

    time_width : scalar, default: 0.5 (500 ms)
        Time bandwidth 'T' (s). Should match data window length.

    freq_width : scalar, default: 4 Hz
        Frequency bandwidth 'W' (Hz)

    n_tapers : scalar, default: (2TW-1)
        Number of tapers to compute. Must be <= 2TW-1, as this is
        the max number of spectrally delimited tapers.

    Returns
    -------
    tapers : ndarray, shape=(n_samples,n_tapers)
        Computed dpss taper functions (n_samples = T*smp_rate)
    """
    # Time-frequency bandwidth product 'TW' (s*Hz)
    time_freq_prod  = time_width*freq_width

    # Up to 2TW-1 tapers are bounded; this is both the default and max value for n_tapers
    n_tapers_max = floor(2*time_freq_prod - 1)
    if n_tapers is None: n_tapers = n_tapers_max

    assert n_tapers <= n_tapers_max, \
        ValueError("For time-freq product = %.1f, %d tapers are tightly bounded in" \
                    "frequency (n_tapers set = %d)" \
                    % (time_freq_prod,n_tapers_max,n_tapers))

    # Convert time bandwidth from s to window length in number of samples
    n_samples = int(round(time_width*smp_rate))

    # Compute the tapers for given window length and time-freq product
    # Note: dpss() normalizes by sum of squares; x sqrt(smp_rate)
    #       converts this to integral of squares (see Chronux function dpsschk())
    # Note: You might imagine you'd want sym=False, but sym=True gives same values
    #       as Chronux dpsschk() function...
    return dpss(n_samples, time_freq_prod, Kmax=n_tapers, sym=True, norm=2).T * sqrt(smp_rate)




# -*- coding: utf-8 -*-
""" Nonparametric permutation/shuffle/randomization statistics """
from warnings import warn
import numpy as np

from spynal.utils import axis_index_slices, isarraylike, one_way_fstat, two_way_fstat
from spynal.randstats.sampling import signs, permutations
from spynal.randstats.utils import resamples_to_pvalue
from spynal.randstats.helpers import _str_to_one_sample_stat, _str_to_assoc_stat, \
                                     _str_to_two_sample_stat, _str_to_one_way_stat, \
                                     _str_to_two_way_stat, _tail_to_compare, \
                                     _paired_sample_data_checks, _two_sample_data_checks


def one_sample_randomization_test(data, axis=0, mu=0, stat='t', tail='both', n_resamples=10000,
                                  seed=None, return_stats=False, keepdims=True, **kwargs):
    """
    Mass univariate 1-sample randomization test

    Parameters and returns are same as :func:`.one_sample_test`

    For each random resample, each observation is randomly assigned a sign
    (+ or -), similar to a Fisher sign test.  The same stat is then computed on
    the resampled data to estimate the distrubution under the null hypothesis,
    and the stat value for the actual observed data is compared to this.

    References
    ----------
    Manly (1997) "Randomization, Bootstrap and Monte Carlo Methods in Biology" ch.6.2
    """
    # Copy data variable to avoid changing values in caller
    data = data.copy()

    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data.ndim + axis

    # Convert string specifiers to callable functions
    stat_func    = _str_to_one_sample_stat(stat,axis)   # Statistic (includes default)
    compare_func = _tail_to_compare(tail)               # Tail-specific comparator

    n = data.shape[axis]
    ndim = data.ndim

    # Subtract hypothetical mean(s) 'mu' from data to center them there
    if mu != 0:  data -= mu

    # Compute statistic of interest on actual observed data
    stat_obs = stat_func(data, **kwargs)

    # Create generators with n_resamples-1 length-n independent Bernoulli RV's
    resamples = signs(n, n_resamples-1, seed)

    if return_stats:
        # Compute statistic under <n_resamples> random resamplings
        stat_shape = [axlen if ax != axis else n_resamples-1 for ax,axlen in enumerate(data.shape)]
        stat_resmp = np.empty(stat_shape)
        for i_resmp,resample in enumerate(resamples):
            # Index into <axis> of data and stat, with ':' for all other axes
            data_slices = axis_index_slices(axis, resample, ndim)
            stat_slices = axis_index_slices(axis, [i_resmp], ndim)
            # Randomly flip signs of obervations flagged by random variables above
            data[data_slices] *= -1
            # Compute statistic on resampled data
            stat_resmp[stat_slices] = stat_func(data, **kwargs)

        # p value = proportion of resampled test statistic values
        #  >= observed value (+ observed value itself)
        p = resamples_to_pvalue(stat_obs, stat_resmp, axis, compare_func)

    else:
        # Compute statistic under <n_resamples> random resamplings and
        # tally values exceeding given criterion (more extreme than observed)
        # Note: Init count to 1's to account for observed value itself
        stat_shape = [axlen if ax != axis else 1 for ax,axlen in enumerate(data.shape)]
        count = np.ones(stat_shape)
        for resample in resamples:
            # Index into <axis> of data and stat, with ':' for all other axes
            data_slices = axis_index_slices(axis, resample, ndim)
            # Randomly flip signs of obervations flagged by random variables above
            data[data_slices] *= -1
            # Compare observed and resampled stats, and tally values passing criterion
            count += compare_func(stat_obs, stat_func(data, **kwargs))

        # p value = proportion of resampled test statistic values passing criterion
        p = count / n_resamples

    # For vector-valued data, extract value from scalar array -> float for output
    if p.size == 1:
        p = p.item()
        if return_stats and isarraylike(stat_obs): stat_obs = stat_obs.item()
    elif not keepdims:
        p = p.squeeze(axis=axis)
        if return_stats: stat_obs = stat_obs.squeeze(axis=axis)

    if return_stats:    return p, stat_obs, stat_resmp
    else:               return p

one_sample_permutation_test = one_sample_randomization_test
""" Alias of :func:`one_sample_randomization_test`. See there for details. """


def paired_sample_permutation_test(data1, data2, axis=0, d=0, stat='t', tail='both',
                                   n_resamples=10000, seed=None, return_stats=False,
                                   keepdims=True, **kwargs):
    """
    Mass univariate paired-sample permutation test

    Parameters and returns are same as :func:`.paired_sample_test`

    For each random resample, each paired-difference observation is randomly
    assigned a sign (+ or -), which is equivalent to randomly permuting each
    pair between samples (data1 vs data2).  The same stat is then computed on
    the resampled data to estimate the distrubution under the null hypothesis,
    and the stat value for the actual observed data is compared to this.

    References
    ----------
    Manly (1997) "Randomization, Bootstrap and Monte Carlo Methods in Biology" ch.6.1
    """
    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data1.ndim + axis

    _paired_sample_data_checks(data1, data2)

    if isinstance(stat,str) and (stat.lower == 'meandiff'): stat = 'mean'

    return one_sample_randomization_test(data1 - data2, axis=axis, mu=d, stat=stat,
                                         tail=tail, n_resamples=n_resamples,
                                         return_stats=return_stats, seed=seed,
                                         keepdims=keepdims, **kwargs)


def paired_sample_association_permutation_test(data1, data2, axis=0, stat='r', tail='both',
                                               n_resamples=10000, seed=None, return_stats=False,
                                               keepdims=True, **kwargs):
    """
    Mass bivariate permutation test of association (eg correlation) between two paired samples

    Parameters and returns are same as :func:`.paired_sample_association_test`

    Observations are randomly permuted across one of the paired samples (data1 vs data2) relative
    to the other, to eliminate any association between them while preserving the marginal
    distributions of each sample.  The same stat is then computed on the resampled data to estimate
    the distrubution under the null hypothesis, and the stat value for the actual observed data
    is compared to this.
    """
    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data1.ndim + axis

    _paired_sample_data_checks(data1, data2)

    # Convert string specifiers to callable functions
    stat_func    = _str_to_assoc_stat(stat,axis)    # Statistic (includes default)
    compare_func = _tail_to_compare(tail)           # Tail-specific comparator

    ndim = data1.ndim
    n = data1.shape[axis]

    # Compute statistic of interest on actual observed data
    stat_obs = stat_func(data1, data2, **kwargs)

    # Create generators with n_resamples-1 random permutations of ints 0:n-1
    resamples = permutations(n, n_resamples-1, seed)

    if return_stats:
        # Compute statistic under <n_resamples> random resamplings
        stat_shape = [axlen if ax != axis else n_resamples-1 for ax,axlen in enumerate(data1.shape)]
        stat_resmp = np.empty(stat_shape)
        for i_resmp,resample in enumerate(resamples):
            # Index into <axis> of data and stat, with ':' for all other axes
            data_slices = axis_index_slices(axis, resample, ndim)
            stat_slices = axis_index_slices(axis, [i_resmp], ndim)
            # Compute statistic on resampled data. Only need to permute one sample
            stat_resmp[stat_slices] = stat_func(data1, data2[data_slices], **kwargs)

        # p value = proportion of resampled test statistic values
        #  >= observed value (+ observed value itself)
        p = resamples_to_pvalue(stat_obs, stat_resmp, axis, compare_func)

    else:
        # Compute statistic under <n_resamples> random resamplings and
        # tally values exceeding given criterion (more extreme than observed)
        # Note: Init count to 1's to account for observed value itself
        stat_shape = [axlen if ax != axis else 1 for ax,axlen in enumerate(data1.shape)]
        count = np.ones(stat_shape)
        for resample in resamples:
            # Index into <axis> of data, with ':' for all other axes
            data_slices = axis_index_slices(axis, resample, ndim)
            # Compute statistic on resampled data1,data2
            stat_resmp  = stat_func(data1, data2[data_slices], **kwargs)
            # Compare observed and resampled stats, and tally values passing criterion
            count += compare_func(stat_obs, stat_resmp)

        # p value = proportion of resampled test statistic values passing criterion
        p = count / n_resamples

    # For vector-valued data, extract value from scalar array -> float for output
    if p.size == 1:
        p = p.item()
        if return_stats and isinstance(stat_obs,np.ndarray): stat_obs = stat_obs.item()
    elif not keepdims:
        p = p.squeeze(axis=axis)
        if return_stats: stat_obs = stat_obs.squeeze(axis=axis)

    if return_stats:    return p, stat_obs, stat_resmp
    else:               return p


def two_sample_permutation_test(data1, data2, axis=0, stat='t', tail='both', n_resamples=10000,
                                 seed=None, return_stats=False, keepdims=True, **kwargs):
    """
    Mass univariate permutation two-sample test

    Parameters and returns are same as :func:`.two_sample_test`

    Observations are permuted across the two samples (data1 vs data2).
    The data is pooled across data1 and and data2, and for each random resample,
    n1 and n2 observations are extracted from the pooled and form the resampled
    versions of data1 and data2, respectively.  The same stat is then computed
    on the resampled data to estimate the distrubution under the null hypothesis,
    and the stat value for the actual observed data is compared to this.

    References
    ----------
    Manly (1997) "Randomization, Bootstrap and Monte Carlo Methods in Biology" ch.6.3
    """
    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data1.ndim + axis

    _two_sample_data_checks(data1, data2, axis)

    # Convert string specifiers to callable functions
    stat_func    = _str_to_two_sample_stat(stat,axis)   # Statistic (includes default)
    compare_func = _tail_to_compare(tail)               # Tail-specific comparator

    ndim = data1.ndim
    n1 = data1.shape[axis]
    n2 = data2.shape[axis]
    N  = n1+n2

    # Compute statistic of interest on actual observed data
    stat_obs = stat_func(data1,data2, **kwargs)

    # Pool two samples together into combined data
    data_pool = np.concatenate((data1,data2),axis=axis)

    # Create generators with n_resamples-1 random permutations of ints 0:N-1
    resamples = permutations(N, n_resamples-1, seed)

    if return_stats:
        # Compute statistic under <n_resamples> random resamplings
        stat_shape = [axlen if ax != axis else n_resamples-1 for ax,axlen in enumerate(data1.shape)]
        stat_resmp = np.empty(stat_shape)
        for i_resmp,resample in enumerate(resamples):
            # Index into <axis> of data and stat, with ':' for all other axes
            # First n1 permuted indexes are resampled "data1"
            # Remaining n2 permuted indexes are resampled "data2"
            data1_slices = axis_index_slices(axis, resample[0:n1], ndim)
            data2_slices = axis_index_slices(axis, resample[n1:], ndim)
            stat_slices = axis_index_slices(axis, [i_resmp], ndim)

            data1_resmp  = data_pool[data1_slices]
            data2_resmp  = data_pool[data2_slices]
            # Compute statistic on resampled data1,data2
            stat_resmp[stat_slices] = stat_func(data1_resmp, data2_resmp, **kwargs)

        # p value = proportion of resampled test statistic values
        #  >= observed value (+ observed value itself)
        p = resamples_to_pvalue(stat_obs, stat_resmp, axis, compare_func)

    else:
        # Compute statistic under <n_resamples> random resamplings and
        # tally values exceeding given criterion (more extreme than observed)
        # Note: Init count to 1's to account for observed value itself
        stat_shape = [axlen if ax != axis else 1 for ax,axlen in enumerate(data1.shape)]
        count = np.ones(stat_shape)
        for resample in resamples:
            # Index into <axis> of data, with ':' for all other axes
            # First n1 permuted indexes are resampled "data1"
            # Remaining n2 permuted indexes are resampled "data2"
            data1_slices = axis_index_slices(axis, resample[0:n1], ndim)
            data2_slices = axis_index_slices(axis, resample[n1:], ndim)

            data1_resmp   = data_pool[data1_slices]
            data2_resmp   = data_pool[data2_slices]
            # Compute statistic on resampled data1,data2
            stat_resmp    = stat_func(data1_resmp, data2_resmp, **kwargs)
            # Compare observed and resampled stats, and tally values passing criterion
            count += compare_func(stat_obs, stat_resmp)

        # p value = proportion of resampled test statistic values passing criterion
        p = count / n_resamples

    # For vector-valued data, extract value from scalar array -> float for output
    if p.size == 1:
        p = p.item()
        if return_stats and isarraylike(stat_obs): stat_obs = stat_obs.item()
    elif not keepdims:
        p = p.squeeze(axis=axis)
        if return_stats: stat_obs = stat_obs.squeeze(axis=axis)

    if return_stats:    return p, stat_obs, stat_resmp
    else:               return p


def one_way_permutation_test(data, labels, axis=0, stat='F', tail='right', groups=None,
                             n_resamples=10000, seed=None, return_stats=False,
                             keepdims=True, **kwargs):
    """
    Mass univariate one-way permutation test

    Parameters and returns are same as :func:`.one_way_test`

    Observation labels are randomly shuffled and allocated to groups
    with the same n's as in the actual observed data.

    References
    ----------
    Manly (1997) "Randomization, Bootstrap and Monte Carlo Methods in Biology" ch.7.1
    """
    if (stat == 'F') and (tail != 'right'):
        warn("For F-test, only right-tailed tests make sense (tail set = %s in args)" % tail)

    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data.ndim + axis

    # Convert string specifiers to callable functions
    stat_func    = _str_to_one_way_stat(stat,axis)  # Statistic (includes default)
    compare_func = _tail_to_compare(tail)           # Tail-specific comparator

    labels  = np.asarray(labels)
    ndim    = data.ndim
    N       = data.shape[axis]

    # Find set of unique group labels in list of labels (if not given)
    if groups is None:
        groups = np.unique(labels)

    # If groups set in args, remove any observations not represented in <groups>
    else:
        idxs = np.in1d(labels, groups)
        if idxs.sum() != N:
            labels  = labels[idxs]
            data    = data[axis_index_slices(axis, idxs, ndim)]
            N       = data.shape[axis]

    # Append <groups> to stat_func args
    if stat_func is one_way_fstat: kwargs.update({'groups':groups})

    # Compute statistic of interest on actual observed data
    stat_obs = stat_func(data, labels, **kwargs)

    # Create generators with n_resamples-1 random permutations of ints 0:N-1
    resamples = permutations(N, n_resamples-1, seed)

    if return_stats:
        # Compute statistic under <n_resamples> random resamplings
        stat_shape = [axlen if ax != axis else n_resamples-1 for ax,axlen in enumerate(data.shape)]
        stat_resmp = np.empty(stat_shape)
        for i_resmp,resample in enumerate(resamples):
            # Index into <axis> of stat, with ':' for all other axes
            stat_slices = axis_index_slices(axis, [i_resmp], ndim)

            # Compute statistic on data and resampled labels
            # Note: Only need to resample labels
            stat_resmp[stat_slices] = stat_func(data, labels[resample], **kwargs)

        # p value = proportion of resampled test statistic values
        #  >= observed value (+ observed value itself)
        p = resamples_to_pvalue(stat_obs, stat_resmp, axis, compare_func)

    else:
        # Compute statistic under <n_resamples> random resamplings and
        # tally values exceeding given criterion (more extreme than observed)
        # Note: Init count to 1's to account for observed value itself
        stat_shape = [axlen if ax != axis else 1 for ax,axlen in enumerate(data.shape)]
        count = np.ones(stat_shape)
        for resample in resamples:
            # Compute statistic on data and resampled labels
            # Note: Only need to resample labels
            stat_resmp = stat_func(data,labels[resample], **kwargs)
            # Compare observed and resampled stats, and tally values passing criterion
            count += compare_func(stat_obs,stat_resmp)

        # p value = proportion of resampled test statistic values passing criterion
        p = count / n_resamples

    # For vector-valued data, extract value from scalar array -> float for output
    if p.size == 1:
        p = p.item()
        if return_stats and isarraylike(stat_obs): stat_obs = stat_obs.item()
    elif not keepdims:
        p = p.squeeze(axis=axis)
        if return_stats: stat_obs = stat_obs.squeeze(axis=axis)

    if return_stats:    return p, stat_obs, stat_resmp
    else:               return p


def two_way_permutation_test(data, labels, axis=0, stat='F', tail='right', groups=None,
                             n_resamples=10000, seed=None, return_stats=False,
                             keepdims=True, **kwargs):
    """
    Mass univariate permutation 2-way test

    Parameters and returns are same as :func:`.two_way_test`

    Observation labels are randomly shuffled and allocated to groups (ANOVA cells)
    with the same n's as in the actual observed data.

    We resample the entire row of labels for each observation, which
    effectively shuffles observations between specific combinations
    of groups/factor levels (ANOVA "cells") cf. recommendation in Manly book

    References
    ----------
    Manly (1997) "Randomization, Bootstrap and Monte Carlo Methods in Biology" ch.7.4
    """
    # todo  Add mechanism to auto-generate interaction term from factors (if interact arg==True)
    if (stat == 'F') and (tail != 'right'):
        warn("For F-test, only right-tailed tests make sense (tail set = %s in args)" % tail)

    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data.ndim + axis

    # Convert string specifier to callable function
    stat_func = _str_to_two_way_stat(stat,axis) # Statistic (includes default))
    compare_func = _tail_to_compare(tail)       # Tail-specific comparator

    labels  = np.asarray(labels)
    n_terms = labels.shape[1]
    N       = data.shape[axis]

    # Find all groups/levels in list of labels (if not given in inputs)
    if groups is None:
        groups = [np.unique(labels[:,term]) for term in range(n_terms)]

    # Append <groups> to stat_func args
    if stat_func is two_way_fstat: kwargs.update({'groups':groups})

    # Compute statistic of interest on actual observed data
    stat_obs = stat_func(data, labels, **kwargs)

    # Create generators with n_resamples-1 random permutations of ints 0:N-1
    resamples = permutations(N, n_resamples-1, seed)

    if return_stats:
        # Compute statistic under <n_resamples> random resamplings
        stat_shape = [axlen if ax != axis else n_terms for ax,axlen in enumerate(data.shape)]
        stat_shape.append(n_resamples-1)
        stat_resmp = np.empty(stat_shape)
        for i_resmp,resample in enumerate(resamples):
            # Compute statistic on data and resampled label rows
            stat_resmp[...,i_resmp] = stat_func(data,labels[resample,:], **kwargs)

        # p value = proportion of resampled test statistic values
        #  >= observed value (+ observed value itself)
        p = resamples_to_pvalue(stat_obs[...,np.newaxis],
                                stat_resmp, -1, compare_func).squeeze(axis=-1)

    else:
        # Compute statistic under <n_resamples> random resamplings and
        # tally values exceeding given criterion (more extreme than observed)
        # Note: Init count to 1's to account for observed value itself
        stat_shape = [axlen if ax != axis else n_terms for ax,axlen in enumerate(data.shape)]
        count = np.ones(stat_shape)
        for resample in resamples:
            # Compute statistic on data and resampled label rows
            stat_resmp = stat_func(data,labels[resample,:], **kwargs)
            # Compare observed and resampled stats, and tally values passing criterion
            count += compare_func(stat_obs,stat_resmp)

        # p value = proportion of resampled test statistic values passing criterion
        p = count / n_resamples

    if return_stats:    return p, stat_obs, stat_resmp
    else:               return p




# -*- coding: utf-8 -*-
""" Band-pass filtering & Hilbert transform-based spectral analysis """
from math import ceil
from collections import OrderedDict
import numpy as np

from scipy.signal import filtfilt, hilbert, zpk2tf, butter, ellip, cheby1, cheby2

from spynal.utils import standardize_array
from spynal.spikes import _spike_data_type, times_to_bool
from spynal.spectra.preprocess import remove_dc
from spynal.spectra.utils import complex_to_spec_type, phase
from spynal.spectra.helpers import _undo_standardize_array_newaxis


def bandfilter_spectrum(data, smp_rate, axis=0, data_type='lfp', spec_type='complex',
                        freqs=((2,8),(10,32),(40,100)), removeDC=True,
                        filt='butter', order=4, params=None, buffer=0, **kwargs):
    """
    Computes band-filtered and Hilbert-transformed signal from data
    for given frequency band(s), then reduces it to 1D frequency spectra by averaging across time.

    Not really the best way to compute 1D frequency spectra, but included for completeness.

    Only parameters differing from :func:`.spectrum` are described here.

    NOTE: Can specify filter implictly using (`freqs`,`filt`,`order`) OR explicitly using `params`.
          If `params` is input, `freqs`, `filt`, and `order` are ignored.

    Parameters
    ----------
    freqs : array-like, shape=(n_freqbands,2), default: ((2,8),(10,32),(40,100))
        List of (low,high) cut frequencies for each band to use.
        Set low cut = 0 for low-pass, set high cut >= smp_rate/2 for high-pass,
        otherwise assumes band-pass. Default samples ~ theta, alpha/beta, gamma.

    filt : str, default: 'butter' (Butterworth)
        Name of filter to use. See :func:`set_filter_params` for all options

    order : int, default: 4
        Filter order

    params : dict
        Parameters that explicitly define filter for each freq band.

        Alternative method for explicitly setting parameters defining freq band filters,
        which are precomputed with :func:`set_filter_params` (or elsewhere).
        Input either `freqs`/`filt`/`order` OR params.
        If params are not explicitly input, we compute them from `freqs`/`filt`/`order`.
        If params *are* explicitly input, `freqs`/`filt`/`order` are ignored.

        One of two forms: 'ba' or 'zpk', with key/values as follows:

        - b,a : array-like, shape=(n_freqbands,) of array-like (n_params[band,])
            Numerator `b` and denominator `a` polynomials of the filter for each band

        - z,p,k :
            Zeros, poles, and system gain of the IIR filter transfer function

    buffer : float, default: 0 (no buffer)
        Time (s) to trim off each end of time dimension of data.
        Removes symmetric buffer previously added (outside of here) to prevent
        edge effects.

    **kwargs :
        Any other kwargs passed directly to :func:`set_filter_params`

    Returns
    -------
    spec : ndarray, shape=(...,n_freqbands,...), dtype=complex or floats
        Band-filtered, (optionally) Hilbert-transformed data, transformed to requested spectral
        type, and averaged across the time axis to 1D frequency spectra.
        Same shape as input data, but with frequency axis replacing time axis.
        dtype is complex if `spec_type` is 'complex', float otherwise.

    freqs : ndarray, shape=(n_freqbands,2)
        List of (low,high) cut frequencies (Hz) for each band used
    """
    if axis < 0: axis = data.ndim + axis

    spec, freqs, _ = bandfilter_spectrogram(data, smp_rate, axis=axis, data_type=data_type,
                                            spec_type=spec_type, freqs=freqs, removeDC=removeDC,
                                            filt=filt, order=order, params=params, buffer=buffer,
                                            **kwargs)

    # Take mean across time axis (which is now shifted +1 b/c of frequency axis)
    return spec.mean(axis=axis+1), freqs


def bandfilter_spectrogram(data, smp_rate, axis=0, data_type='lfp', spec_type='complex',
                           freqs=((2,8),(10,32),(40,100)), removeDC=True,
                           filt='butter', order=4, params=None, buffer=0, downsmp=1, **kwargs):
    """
    Computes zero-phase band-filtered and Hilbert-transformed signal from data
    for given frequency band(s).

    Function aliased as bandfilter().

    Only parameters differing from :func:`.spectrogram` are described here.

    NOTE: Can specify filter implictly using (`freqs`,`filt`,`order`) OR explicitly using `params`.
          If `params` is input, `freqs`, `filt`, and `order` are ignored.

    Parameters
    ----------
    freqs : array-like, shape=(n_freqbands,2), default: ((2,8),(10,32),(40,100))
        List of (low,high) cut frequencies for each band to use.
        Set low cut = 0 for low-pass, set high cut >= smp_rate/2 for high-pass,
        otherwise assumes band-pass. Default samples ~ theta, alpha/beta, gamma.

    filt : str, default: 'butter' (Butterworth)
        Name of filter to use. See :func:`set_filter_params` for all options

    order : int, default: 4
        Filter order

    params : dict, default: (computed from `freqs`/`filt`/`order`)
        Parameters that explicitly define filter for each freq band.

        Alternative method for explicitly setting parameters defining freq band filters,
        which are precomputed with :func:`set_filter_params` (or elsewhere).
        Input either `freqs`/`filt`/`order` OR params.
        If params are not explicitly input, we compute them from `freqs`/`filt`/`order`.
        If params *are* explicitly input, `freqs`/`filt`/`order` are ignored.

        One of two forms: 'ba' or 'zpk', with key/values as follows:

        - b,a : array-like, shape=(n_freqbands,) of array-like (n_params[band,])
            Numerator `b` and denominator `a` polynomials of the filter for each band

        - z,p,k :
            Zeros, poles, and system gain of the IIR filter transfer function

    buffer : float, default: 0 (no buffer)
        Time (s) to trim off each end of time dimension of data.
        Removes symmetric buffer previously added (outside of here) to prevent
        edge effects.

    **kwargs :
        Any other kwargs passed directly to :func:`set_filter_params`

    Returns
    -------
    spec : ndarray, shape=(...,n_freqbands,n_timepts_out,...), dtype=complex or float.
        Band-filtered, (optionally) Hilbert-transformed "spectrogram" of data,
        transformed to requested spectral type.
        Same shape as input data, but with frequency axis prepended immediately
        before time `axis`. dtype is complex if `spec_type` is 'complex', float otherwise.

    freqs : ndarray, shape=(n_freqbands,2)
        List of (low,high) cut frequencies (Hz) for each band used in `spec`.

    timepts : ndarray, shape=(n_timepts_out,)
        List of timepoints in `spec` (in s, referenced to start of data).
    """
    if axis < 0: axis = data.ndim + axis

    # Convert spike timestamp data to boolean spike train format
    if (data_type == 'spike') and (_spike_data_type(data) == 'timestamp'):
        lims    = kwargs.pop('lims',None)
        bins    = kwargs.pop('bins',None)
        data,_  = times_to_bool(data, width=1/smp_rate, lims=lims, bins=bins)
        axis    = data.ndim

    # Convert buffer from s -> samples
    if buffer != 0:  buffer  = int(ceil(buffer*smp_rate))

    # Set filter parameters from frequency bands if <params> not explicitly passed in
    if params is None:
        assert freqs is not None, \
            ValueError("Must input a value for either filter <params> or band <freqs>")

        freqs   = np.asarray(freqs)  # Convert freqs to (n_freqbands,2)
        n_freqs = freqs.shape[0]
        params  = set_filter_params(freqs, smp_rate, filt=filt, order=order,
                                    form='ba', return_dict=True, **kwargs)

    # Determine form of filter parameters given: b,a or z,p,k
    else:
        assert len(kwargs) == 0, \
            TypeError("Incorrect or misspelled variable(s) in keyword args: " +
                      ', '.join(kwargs.keys()))

        if np.all([(param in params) for param in ['b','a']]):       form = 'ba'
        elif np.all([(param in params) for param in ['z','p','k']]): form = 'zpk'
        else:
            raise ValueError("<params> must be a dict with keys 'a','b' or 'z','p','k'")

        # Convert zpk form to ba
        if form == 'zpk':
            n_freqs = len(params['z'])
            params['b'] = [None] * n_freqs
            params['a'] = [None] * n_freqs
            for i_freq in range(n_freqs):
                b,a = zpk2tf(params['z'][i_freq],params['p'][i_freq],params['k'][i_freq])
                params['b'][i_freq] = b
                params['a'][i_freq] = a
        else:
            n_freqs = len(params['b'])

    # Set any freqs > Nyquist equal to Nyquist
    if freqs is not None: freqs[freqs > smp_rate/2] = smp_rate/2

    # Reshape data array -> (n_timepts_in,n_dataseries) matrix
    data, data_shape = standardize_array(data, axis=axis, target_axis=0)
    # Temporarily append singleton axis to vector-valued data to simplify code
    vector_data = data.ndim == 1
    if vector_data: data = data[:,np.newaxis]

    n_timepts_in,n_series = data.shape

    # Time indexes to extract from spectrogram for output (accounting for buffer, downsampling)
    time_idxs_out   = np.arange(buffer,n_timepts_in-buffer,downsmp)
    n_timepts_out   = len(time_idxs_out)

    if removeDC: data = remove_dc(data,axis=0)

    dtype = float if spec_type == 'real' else complex
    spec = np.empty((n_freqs,n_timepts_out,n_series),dtype=dtype)

    # For each frequency band, band-filter raw signal and
    # compute complex analytic signal using Hilbert transform
    for i_freq,(b,a) in enumerate(zip(params['b'],params['a'])):
        bandfilt = filtfilt(b, a, data, axis=0, method='gust')
        # Note: skip Hilbert transform for real output
        spec[i_freq,:,:] = bandfilt[time_idxs_out,:] if spec_type == 'real' else \
                           hilbert(bandfilt[time_idxs_out,:],axis=0)

    # Convert to desired output spectral signal type
    spec    = complex_to_spec_type(spec,spec_type)

    if vector_data: spec = spec.squeeze(axis=-1)
    spec = _undo_standardize_array_newaxis(spec,data_shape,axis=axis)

    timepts = time_idxs_out.astype(float)/smp_rate  # Convert time sampling from samples -> s

    return spec, freqs, timepts

bandfilter = bandfilter_spectrogram
""" Alias of :func:`bandfilter_spectrogram`. See there for details. """


def set_filter_params(bands, smp_rate, filt='butter', order=4, form='ba',
                      return_dict=False, **kwargs):
    """
    Sets coefficients for desired filter(s) using scipy.signal
    "Matlab-style IIR filter design" functions

    NOTE: If return_dict is False, outputs are returned as a tuple, as described below;
    else, outputs are packaged in a single dict, with param names as keys.

    Parameters
    ----------
    bands : array-like, shape=(n_freqbands,2)
        List of (low,high) cut frequencies for each band to use.
        Set low cut = 0 for low-pass, set high cut >= smp_rate/2 for high-pass,
        otherwise assumes band-pass

    smp_rate : scalar
        Data sampling rate (Hz)

    filt : str, default: 'butter' (Butterworth)
        Name of filter to use. See :func:`set_filter_params` for all options

    order : int, default: 4
        Filter order

    form : {'ba','zpk'}, default: ‘ba’
        Type of parameters to output:
        - 'ba': numerator(b), denominator (a)
        - 'zpk': Zeros (z), poles (p), and system gain (k) of the IIR filter transfer function

    return_dict : bool, default: False
        If True, params returned in a dict; else as standard series (tuple) of output variables

    **kwargs :
        Any additional kwargs passed directly to filter function

    Returns
    -------
    b,a : list, shape=(n_freqbands,) of list, shape=(n_params[band,])
        Numerator `b` and denominator `a` polynomials of the filter for each band.
        Returned if `form` == 'ba'.

    z,p,k : list, shape=(n_freqbands,) of list, shape=(n_params[band,])
        Zeros, poles, and system gain of IIR transfer function.
        Returned if `form` == 'zpk'.

    Examples
    --------
    params = set_filter_params(bands, smp_rate, form='ba', return_dict=True)

    b,a = set_filter_params(bands, smp_rate, form='ba', return_dict=False)

    z,p,k = set_filter_params(bands, smp_rate, form='zpk', return_dict=False)
    """
    # Convert bands to (n_freqbands,2)
    bands       = np.asarray(bands)
    if bands.ndim == 1: bands = np.reshape(bands,(1,len(bands)))
    n_bands     = bands.shape[0]
    nyquist     = smp_rate/2.0   # Nyquist freq at given sampling freq

    # Setup filter-generating function for requested filter type
    # Butterworth filter
    if filt.lower() in ['butter','butterworth']:
        gen_filt = lambda band,btype: butter(order,band,btype=btype,output=form)
    # Elliptic filter
    elif filt.lower() in ['ellip','butterworth']:
        rp = kwargs.pop('rp',5)
        rs = kwargs.pop('rs',40)
        gen_filt = lambda band,btype: ellip(order,rp,rs,band,btype=btype,output=form)
    # Chebyshev Type 1 filter
    elif filt.lower() in ['cheby1','cheby','chebyshev1','chebyshev']:
        rp = kwargs.pop('rp',5)
        gen_filt = lambda band,btype: cheby1(order,rp,band,btype=btype,output=form)
    # Chebyshev Type 2 filter
    elif filt.lower() in ['cheby2','chebyshev2']:
        rs = kwargs.pop('rs',40)
        gen_filt = lambda band,btype: cheby2(order,rs,band,btype=btype,output=form)
    else:
        raise ValueError("Filter type '%s' is not supported (yet)" % filt)
    assert len(kwargs) == 0, \
        TypeError("Incorrect or misspelled variable(s) in keyword args: "+', '.join(kwargs.keys()))

    # Setup empty lists to hold filter parameters
    if form == 'ba':    params = OrderedDict({'b':[None]*n_bands, 'a':[None]*n_bands})
    elif form == 'zpk': params = OrderedDict({'z':[None]*n_bands, 'p':[None]*n_bands,
                                              'k':[None]*n_bands})
    else:
        raise ValueError("Output form '%s' is not supported. Should be 'ba' or 'zpk'" % form)

    for i_band,band in enumerate(bands):
        band_norm = band/nyquist  # Convert band to normalized frequency

        # If low-cut freq = 0, assume low-pass filter
        if band_norm[0] == 0:   btype = 'lowpass';  band_norm = band_norm[1]
        # If high-cut freq >= Nyquist freq, assume high-pass filter
        elif band_norm[1] >= 1: btype = 'highpass'; band_norm = band_norm[0]
        # Otherwise, assume band-pass filter
        else:                   btype = 'bandpass'

        if form == 'ba':
            params['b'][i_band],params['a'][i_band] = gen_filt(band_norm,btype)
        else:
            params['z'][i_band],params['p'][i_band],params['k'][i_band] = gen_filt(band_norm,btype)

    if return_dict: return params
    else:           return params.values()




from spynal.sync.sync import synchrony, spike_field_coupling, simulate_multichannel_oscillation
from spynal.sync.coherence import coherence, spike_field_coherence, ztransform_coherence
from spynal.sync.phasesync import plv, phase_locking_value, ppc, pairwise_phase_consistency, \
                                  spike_field_plv, spike_field_phase_locking_value, \
                                  spike_field_ppc, spike_field_pairwise_phase_consistency, \
                                  plv_to_ppc



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Functions for loading from and saving to Matlab v7 (and earlier) MAT files using scipy.io """
import numpy as np

import scipy.io

from spynal.matIO.helpers import _v7_matlab_type, _process_v7_object, \
                                 _is_structured_array, _structuredarray_to_dict, \
                                 _structuredarray_to_dataframe, DEBUG



def _load7(filename, variables=None, typemap=None, extract_items=None, order='Matlab'):
    """
    Load data variables from a version 7 (or older) Matlab MAT file
    Uses scipy.io.loadmat to load data
    """
    # scipy.io returns arrays in column-major (Matlab/Fortran) order by default
    # Transpose/permute array axes if Python/C/row-major requested
    transpose = order.upper() in ['PYTHON','C','ROW','ROWMAJOR']

    # If <variables> is set, convert to {variable_name:None} dict
    # (otherwise leave = None, so it loads all variables from datafile)
    if variables is not None:
        variables = {vbl:None for vbl in variables}

    # Load each requested variable and save each as Numpy array in a {variable_name:array} dict
    data = scipy.io.loadmat(filename,variable_names=variables)

    # Get rid of header info returned by loadmat() (variables starting with '_...')
    data = {vbl:value for vbl,value in data.items() if vbl[0] != '_'}

    # Ensure we actually loaded all requested variables
    if variables is not None:
        for vbl in variables:
            assert vbl in data.keys(), \
                ValueError("Variable '%s' not present in file %s" % (vbl,filename))

    # If <variables> not set, load all variables in datafile
    else:
        variables = list(data.keys())

    # Get data into desired output format
    for vbl in variables:
        # Infer Matlab variable type from object structure
        matlab_vbl_type = _v7_matlab_type(data[vbl])
        # If specific variable name is listed in <typemap>, use given mapping
        python_vbl_type = typemap[vbl] if vbl in typemap else None

        # If specific variable name is listed in <extract_items>, use associated value
        if vbl in extract_items:                extract_item = extract_items[vbl]
        # Otherwise, if variable type is listed in <extract_items>, use associated value
        elif matlab_vbl_type in extract_items:  extract_item = extract_items[matlab_vbl_type]
        # Otherwise, just use value for 'array' as generic value for all other (eg scalar) types        
        else:                                   extract_item = extract_items['array']
        if DEBUG: print("'%s'" % vbl, matlab_vbl_type, python_vbl_type, extract_item)

        # Process loaded object -- extract data, convert to appropriate Python type,
        # traversing down into object (cell elements/struct fields) with recursive calls as needed
        data[vbl] = _process_v7_object(data[vbl], matlab_vbl_type=matlab_vbl_type,
                                       python_vbl_type=python_vbl_type, extract_item=extract_item,
                                       typemap=typemap, extract_items=extract_items,
                                       transpose=transpose, level=0, vbl=vbl)

    return data


def _who7(filename, **kwargs):
    """  List data variables from a version 7 (or older) Matlab MAT file """
    # Load list of 3-tuples of (variable,size,type) for each variable in file
    variables = scipy.io.whosmat(filename, appendmat=True, **kwargs)
    # Extract and return just the variable names
    return [vbl[0] for vbl in variables]


def _save7(filename, variables, **kwargs):
    """
    Save data variables to version 7 Matlab MAT file
    Uses scipy.io.savemat to save data
    """
    scipy.io.savemat(filename,variables,**kwargs)




from spynal.randstats.randstats import one_sample_test, \
                                       paired_sample_test, paired_sample_test_labels, \
                                       paired_sample_association_test, paired_sample_association_test_labels, \
                                       two_sample_test, two_sample_test_labels, \
                                       one_way_test, two_way_test, \
                                       one_sample_confints, paired_sample_confints, two_sample_confints
from spynal.randstats.sampling import permutations, bootstraps, signs, jackknifes, subsets
from spynal.randstats.permutation import one_sample_randomization_test, one_sample_permutation_test, \
                                         paired_sample_permutation_test, paired_sample_association_permutation_test, \
                                         two_sample_permutation_test, \
                                         one_way_permutation_test, two_way_permutation_test
from spynal.randstats.bootstrap import one_sample_bootstrap_test, paired_sample_bootstrap_test, \
                                       paired_sample_association_bootstrap_test, two_sample_bootstrap_test
from spynal.randstats.utils import resamples_to_pvalue, confint_to_indexes, jackknife_to_pseudoval



# -*- coding: utf-8 -*-
""" Helper functions for spectra module for spectral analysis """
from warnings import warn
from math import floor, ceil, log2
import numpy as np

from multiprocessing import cpu_count
from pyfftw.interfaces.scipy_fftpack import fft, ifft # ~ 46/16 s on benchmark

# Set default arguments for pyfftw functions: Fast planning, use all available threads
_FFTW_KWARGS_DEFAULT = {'planner_effort': 'FFTW_ESTIMATE',
                        'threads': cpu_count()}

# from numpy.fft import fft,ifft        # ~ 15 s on benchmark
# from scipy.fftpack import fft,ifft    # ~ 11 s on benchmark
# from mkl_fft import fft,ifft    # ~ 15.2 s on benchmark
# from pyfftw import empty_aligned, byte_align
# from pyfftw.interfaces.cache import enable as enable_pyfftw_cache
# import pyfft
# enable_pyfftw_cache()


def _extract_triggered_data(data, smp_rate, event_times, window):
    """
    Extracts windowed chunks of data around given set of event times

    Parameters
    ----------
    data : ndarray, shape=(n_samples,...)
        Data to cut triggered snippets out of
        NOTE: Not coded up for arbitrary-shaped data

    smp_rate : int, default: 1000
        Sampling rate for `data` (Hz)

    event_times : array-like, shape=(n_events,)
        List of times (s) of event triggers to extract data around.
        Times are referenced to 1st data sample (t=0).

    window : array-like, shape=(2,)
        [start,end] of window (in s) to extract around each trigger

    Returns
    -------
    data : ndarray, shape=(n_samples_per_window,n_events,...)
        Data cut at event triggers
    """
    # Convert event_times, window from s -> samples
    event_times = np.floor(np.asarray(event_times)*smp_rate).astype(int)
    window      = np.round(np.asarray(window)*smp_rate).astype(int)

    n_per_event = window[1] - window[0]
    n_events    = len(event_times)
    data_shape  = data.shape
    data_out    = np.zeros((n_per_event,n_events,*data_shape[1:]))

    for i_event,event in enumerate(event_times):
        idxs    = np.arange(event-window[0],event+window[1])
        data_out[:,i_event,...] = data[idxs,...]

    return data_out


def _undo_standardize_array_newaxis(data,data_shape,axis=0):
    """
    Reshape data array from unwrapped form back to ~ original
    multi-dimensional form in special case where a new frequency axis was
    inserted before time axis (<axis>)

    Parameters
    ----------
    data : ndarray, shape=(axis_len,m)
        Data array w/ all axes > 0 unwrapped into single dimension, where m = prod(shape[1:])

    data_shape : tuple, len=Any
        Original shape of data array

    axis : int, default: 0
        Axis of original data corresponding to distinct observations,
        which has become axis 1, but will be permuted back to original axis.

    Returns
    -------
    data: ndarray, shape=(...,n_freqs,n_timepts,...)
        Data array reshaped back to original shape
    """
    data_shape  = np.asarray(data_shape)
    if axis < 0: axis = len(data_shape) + axis

    data_ndim   = len(data_shape) # Number of dimensions in original data
    n_freqs      = data.shape[0]
    n_timepts    = data.shape[1]

    # If data array data had > 2 dims, reshape matrix back into ~ original shape
    if len(data_shape) > 2:
        shape   = (n_freqs, n_timepts, *data_shape[np.arange(data_ndim) != axis])
        data    = np.reshape(data,shape,order='F')

    # Squeeze (n,1) array back down to 1d (n,) vector,
    #  and extract value from scalar array -> float
    elif data_ndim == 1:
        data = data.squeeze(axis=-1)
        if data.size == 1: data = data.item()

    # If <axis> wasn't 0, move axis back to original position
    if (axis != 0) and isinstance(data,np.ndarray):
        data = np.moveaxis(data,(0,1),(axis,axis+1))

    return data


def _infer_freq_scale(freqs):
    """ Determine if frequency sampling vector is linear, logarithmic, or uneven """
    # Determine if frequency scale is linear (all equally spaced)
    if np.allclose(np.diff(np.diff(freqs)),0):
        return 'linear'

    # Determine if frequency scale is logarithmic (all equally spaced in log domain)
    elif np.allclose(np.diff(np.diff(np.log2(freqs))),0):
        return 'log'

    # Otherwise assume arbitrary unevenly-sampled frequency axis (as in bandfilter/burst analysis)
    else:
        warn("Unable to determine scale of frequency sampling vector. Assuming it's arbitrary")
        return 'uneven'


def _frequency_plot_settings(freqs):
    """ Return settings for plotting a frequency axis: plot freqs, ticks, tick labels """
    freqs = np.asarray(freqs).squeeze()
    # For freqs given as (low,high) bands, convert to band means
    if (freqs.ndim == 2) and (freqs.shape[1] == 2): freqs = freqs.mean(axis=1)

    freq_scale = _infer_freq_scale(freqs)

    # For log-sampled freqs, plot in log2(freq) but label with actual freqs
    if freq_scale == 'log':
        freqs           = np.log2(freqs)            # Log2-transform plotting freqs
        fmin            = ceil(freqs[0])
        fmax            = floor(freqs[-1])
        freq_ticks      = np.arange(fmin,fmax+1)    # Plot ticks every octave: [2,4,8,16,...]
        freq_tick_labels= 2**np.arange(fmin,fmax+1)

    # For linear-sampled freqs, just plot in actual freqs
    elif freq_scale == 'linear':
        fmin            = ceil(freqs[0]/10.0)*10.0  # Plot ticks every 10 Hz
        fmax            = floor(freqs[-1]/10.0)*10.0
        freq_ticks      = np.arange(fmin,fmax+1,10).astype(int)
        freq_tick_labels= freq_ticks

    # For arbitrary unevenly-sampled freqs (eg bandfilter or burst analyis),
    # plot freqs categorically as range 0 - n_freqs-1, but label with actual freqs
    else:
        freq_tick_labels= freqs
        freq_ticks      = np.arange(len(freqs))
        freqs           = np.arange(len(freqs))

    return freqs, freq_ticks, freq_tick_labels


def _str_to_pool_func(func):
    """ Convert string specifier to callable pooling function """
    # If it's already a callable, return as-is
    if callable(func): return func

    assert isinstance(func,str), \
        TypeError("Unsupported type '%s' for <func>. Use string or function" % type(func))

    func = func.lower()
    if func == 'mean':  return lambda x: np.mean(x, axis=0)
    elif func == 'sum': return lambda x: np.sum(x, axis=0)
    else:
        raise ValueError("Unsupported value '%s' for func. Set='mean'|'sum'" % func)



# -*- coding: utf-8 -*-
""" Phase-based oscillatory neural synchrony analysis """
from warnings import warn
import numpy as np

from spynal.utils import axis_index_slices, setup_sliding_windows
from spynal.randstats.sampling import jackknifes
from spynal.randstats.utils import jackknife_to_pseudoval
from spynal.sync.helpers import _sync_raw_to_spectral, _sfc_raw_to_spectral


# =============================================================================
# Field-Field Synchrony functions
# =============================================================================
def plv(data1, data2, axis=0, return_phase=False, transform=None, single_trial=None,
        spec_method='wavelet', data_type=None, smp_rate=None,
        time_axis=None, taper_axis=None, keepdims=True, **kwargs):
    """
    Compute phase locking value (PLV) between raw or spectral (time-frequency) LFP data.

    PLV is a measure of phase synchrony that ignores signals amplitudes.

    PLV is the mean resultant length (magnitude of the vector mean) of phase
    differences dphi btwn phases of data1 and data2::

        dphi = phase(data1) - phase(data2)
        PLV  = abs( trial_mean(exp(i*dphi)) )

    Only parameters differing from :func:`synchrony` are described here.

    Parameters
    ----------
    transform : 'PPC' or None, default: None
        Transform to apply to all computed PLV values.
        Set=None to return untransformed PLV.
        Set='PPC' to transform to debiased estimator of PLV^2 (aka Pairwise Phase Consistency/PPC).

    **kwargs :
        Any other keyword args passed as-is to spectrogram() function.

    References
    ----------
    Lachaux (1999) https://doi.org/10.1002/(SICI)1097-0193(1999)8:4%3C194::AID-HBM4%3E3.0.CO;2-C
    """
    assert data1.shape[axis] == data2.shape[axis], \
        ValueError("data1,data2 must have same number of observations (trials)")
    assert not((single_trial is not None) and return_phase), \
        ValueError("Cannot do both single_trial AND return_phase together")

    if axis < 0: axis = data1.ndim + axis
    if (time_axis is not None) and (time_axis < 0): time_axis = data1.ndim + time_axis
    if (taper_axis is not None) and (taper_axis < 0): taper_axis = data1.ndim + taper_axis

    # Check if raw data input. If so, compute spectral transform first
    data1, data2, freqs, timepts, axis, time_axis, taper_axis = \
        _sync_raw_to_spectral(data1, data2, smp_rate, axis, time_axis, taper_axis,
                              spec_method, data_type, **kwargs)

    # For multitaper, compute means across trials, tapers; otherwise just do means across trials
    reduce_axes = (axis,taper_axis) if spec_method == 'multitaper' else axis
    n = np.prod([data1.shape[ax] for ax in reduce_axes]) if spec_method == 'multitaper' else \
        data1.shape[axis]

    # Setup actual function call for any transform to compute on raw PLV values
    if (transform is None) or callable(transform):
        pass
    elif isinstance(transform,str):
        transform = transform.lower()
        if transform == 'ppc':
            transform = lambda PLV: plv_to_ppc(PLV, n)
        else:
            raise ValueError("Unsupported value '%s' set for <transform>" % transform)
    else:
        raise TypeError("Unsupported type '%s' for <transform>. Use string or function or None" \
                        % type(transform))

    # Compute normalized cross-spectrum btwn the two channels
    # Cross-spectrum-based method adapted from FieldTrip ft_conectivity_ppc()
    # Note: Traditional circular mean-based algorithm is ~3x slower
    cross_spec = data1*data2.conj()                 # Compute cross-spectrum
    cross_spec = cross_spec / np.abs(cross_spec)    # Normalize cross-spectrum


    def _cross_to_plv(cross_spec, axis, return_phase, transform, keepdims):
        """ Compute PLV from cross-spectrum """
        # Vector-average complex spectra across observatations (trials and/or tapers)
        vector_mean = np.mean(cross_spec, axis=axis, keepdims=keepdims)

        # Compute PLV = magnitude of complex vector mean = mean resultant length
        # Note: .real deals with floating point error, converts complex dtypes to float
        PLV = np.abs(vector_mean).real

        # Perform any requested tranform on PLV (eg convert to PPC)
        if transform is not None: PLV = transform(PLV)

        # Optionally also extract mean relative phase angle
        if return_phase:    return PLV, np.angle(vector_mean)
        else:               return PLV, None


    # Standard across-trial PLV estimator
    if single_trial is None:
        PLV, dphi = _cross_to_plv(cross_spec, reduce_axes, return_phase, transform, keepdims)

    # Single-trial PLV estimator using jackknife resampling method
    else:
        # Jackknife resampling of PLV statistic (this is the 'richter' estimator)
        # Note: Allow for reduction along taper axis within resampled stat function, but only
        #       resample across trial axis--want sync estimates for single-trials, not tapers
        jackfunc = lambda s12: _cross_to_plv(s12, reduce_axes, False, transform, True)[0]
        plv_shape = list(data1.shape)
        if spec_method == 'multitaper': plv_shape[taper_axis] = 1
        n_jack = plv_shape[axis]
        ndim = len(plv_shape)

        # Create generator with n length-n vectors, each of which excludes 1 trial
        resamples = jackknifes(n_jack)

        # Do jackknife resampling -- estimate statistic w/ each observation left out
        PLV = np.empty(plv_shape, dtype=float)
        for trial,sel in enumerate(resamples):
            # Index into <axis> of data and stat, with ':' for all other axes
            slices_in   = axis_index_slices(axis, sel, ndim)
            slices_out  = axis_index_slices(axis, [trial], ndim)
            PLV[slices_out] = jackfunc(cross_spec[slices_in])

        # Convert to jackknife pseudovalues = n*stat_full - (n-1)*stat_jackknife
        # Note: n here is the number of jackknifes computed = n_trials
        if single_trial == 'pseudo':
            plv_full = jackfunc(cross_spec)
            PLV = jackknife_to_pseudoval(plv_full, PLV, data1.shape[axis])

        if not keepdims and (spec_method == 'multitaper'): PLV = PLV.squeeze(axis=taper_axis)

    if return_phase:    return  PLV, freqs, timepts, dphi
    else:               return  PLV, freqs, timepts

# Alias function with full name
phase_locking_value = plv
""" Alias of :func:`plv`. See there for details. """


def ppc(data1, data2, axis=0, return_phase=False, single_trial=None,
        spec_method='wavelet', data_type=None, smp_rate=None,
        time_axis=None, taper_axis=None, keepdims=True, **kwargs):
    """
    Compute pairwise phase consistency (PPC) between raw or spectral
    (time-frequency) LFP data, which is bias-corrected (unlike PLV and coherence,
    which are biased by n)

    PPC is a measure of phase synchrony that ignores signal amplitude.

    PPC computes the cosine of the absolute angular distance (the vector dot product)
    for all given pairs of relative phases, i.e., it computes how similar the relative
    phase observed in one trial is to the relative phase observed in another trial

    PPC is also an debiased estimator of the square of the mean resultant length (PLV^2),
    and can be expressed (and computed efficiently) in terms of PLV and n::

        PPC = (n*PLV^2 - 1) / (n-1)

    Only parameters differing from :func:`.synchrony` are described here.

    Parameters
    ----------
    **kwargs :
        Any other keyword args passed as-is to spectrogram() function.

    References
    ----------
    - Original concept:   Vinck et al. (2010) https://doi.org/10.1016/j.neuroimage.2010.01.073
    - Relation to PLV:    Kornblith, Buschman, Miller (2015) https://doi.org/10.1093/cercor/bhv182
    """
    # Simply call plv() with a 'PPC' transform
    return plv(data1, data2, axis=axis, return_phase=return_phase, transform='PPC',
               single_trial=single_trial, spec_method=spec_method, data_type=data_type,
               smp_rate=smp_rate,  time_axis=time_axis, taper_axis=taper_axis,
               keepdims=keepdims, **kwargs)

# Alias function with full name
pairwise_phase_consistency = ppc
""" Alias of :func:`ppc`. See there for details. """


# =============================================================================
# Spike-Field Synchrony functions
# =============================================================================
def spike_field_plv(spkdata, lfpdata, axis=0, time_axis=None, taper_axis=None,
                    timepts=None, width=0.5, spacing=None, lims=None, timewins=None,
                    data_type=None, spec_method='wavelet', smp_rate=None,
                    return_phase=False, keepdims=True, **kwargs):
    """
    Compute phase locking value (PLV) of spike-triggered LFP phase

    PLV is the mean resultant length (magnitude of the vector mean) of the
    spike-triggered LFP phase 'phi'::

        PLV  = abs( trial_mean(exp(i*phi)) )

    Because spiking response are sparse, spike-LFP PLV is typically computed within sliding
    time windows (ie summation across trials AND across within-window timepoints). These can
    be specified either explicitly using `timewins` or implicitly using `width`/`spacing`/`lims`.

    Only parameters differing from :func:`.spike_field_coupling` are described here.

    Parameters
    ----------
    width : scalar, default: 0.5 (500 ms)
        Width (in s) of sliding time windows for computing PLV

    spacing : scalar, default: set = `width` (ie exactly non-overlapping windows)
        Spacing (in s) of sliding time windows for computing PLV

    lims : array-like, shape=(2,), default: (timepts[0],timepts[-1]) (full sampled time of data)
        [Start,end] time limits (in s) for full series of sliding windows

    timewins : ndarray, shape=(n_timewins,2), default: setup_sliding_windows(width,lims,spacing)
        Alternative method for setting sliding time windows; overrides `width`/`spacing`/`lims`.
        Custom time windows to compute PLV within, given as explicit series of window [start,end]'s
        (in s). Can have any arbitrary width/spacing (eg to compute PLV in arbitary time epochs).
        Default generates sliding windows with `width` and `spacing` from `lims[0]` to `lims[1]`

        Special case: Set = (lim[0],lim[1]) to compute PLV *spectrum* over entire data time period.

    **kwargs :
        Any other keyword args passed as-is to spectrogram() function.

    References
    ----------
    Lachaux (1999) https://doi.org/10.1002/(SICI)1097-0193(1999)8:4%3C194::AID-HBM4%3E3.0.CO;2-C
    """
    max_axis_mismatch = 2 if spec_method == 'multitaper' else 1
    assert (spkdata.ndim == lfpdata.ndim) and \
           ((np.array(spkdata.shape) != np.array(lfpdata.shape)).sum() <= max_axis_mismatch), \
        ValueError("Spiking data " + str(spkdata.shape) +
                   " must have same size/shape as LFP data " + str(lfpdata.shape) +
                   " (w/ singleton to match freq [and taper] axis)")
    if (timepts is None) and ((timewins is not None) or (lims is not None)):
        assert smp_rate is not None, \
            ValueError("If no value is input for <timepts>, must input value for <smp_rate>")
        warn("No value input for <timepts>. Setting = (0 - n_timepts-1)/smp_rate.\n"
             "Assuming <lims> and/or <timewins> are given in same timebase.")

    # if data_type is None: data_type = _infer_data_type(lfpdata)
    if axis < 0: axis = lfpdata.ndim + axis
    if time_axis < 0: time_axis = lfpdata.ndim + time_axis

    # Default timepts to range from 0 - n_timepts/smp_rate
    if timepts is None:     timepts = np.arange(lfpdata.shape[time_axis]) / smp_rate
    elif smp_rate is None:  smp_rate = 1 / np.diff(timepts).mean()

    # Are we computing PLV as a first step for PPC? (should only be set when called from ppc())
    _for_ppc = kwargs.pop('_for_ppc', False)

    # Check if raw data input. If so, compute spectral transform first
    spkdata, lfpdata, freqs, timepts, smp_rate, axis, time_axis, taper_axis = \
        _sfc_raw_to_spectral(spkdata, lfpdata, smp_rate, axis, time_axis, taper_axis, timepts,
                             'plv', spec_method, data_type, **kwargs)

    # Default lims to [start,end] of timepts
    # (Note: do this after multitaper timepts adjustment above)
    if lims is None: lims = (timepts[0],timepts[-1])

    # For multitaper spectral data, reshape lfpdata s.t. tapers and trials are on same axis
    if spec_method == 'multitaper':
        assert taper_axis is not None, \
            ValueError("For multitaper spec_method, must input a value for taper_axis")

        n_tapers = lfpdata.shape[taper_axis]

        # Move taper axis next to trial axis
        lfpdata = np.moveaxis(lfpdata,taper_axis,axis)
        # If trial axis was after taper axis, taper axis is now after trial, so unwrap in F order
        # If trial axis was before taper axis, taper axis is now before trial, so unwrap in C order
        order = 'F' if axis > taper_axis else 'C'
        axis_ = axis - 1 if axis > taper_axis else axis
        # Reshape lfpdata so tapers on same axis as trials -> (...,n_trials*n_tapers,...)
        lfpdata = lfpdata.reshape((*lfpdata.shape[0:axis_], -1, *lfpdata.shape[(axis_+2):]),
                                  order=order)

        # Expand spkdata trial axis to n_trials*n_tapers to match lfpdata and remove taper axis
        tiler   = np.ones((spkdata.ndim,),dtype=int)
        tiler[axis] = n_tapers
        spkdata = np.tile(spkdata, tuple(tiler))
        spkdata = spkdata.squeeze(axis=taper_axis)

        # Adjust axes for removal of taper axis
        if time_axis > taper_axis: time_axis -= 1
        if axis > taper_axis: axis -= 1

    data_ndim = lfpdata.ndim
    # Move time and observations/trials axes to end of arrays -> (...,n_timepts,n_obs)
    if not ((time_axis == data_ndim-2) and (axis == data_ndim-1)):
        lfpdata = np.moveaxis(lfpdata,time_axis,-1)
        lfpdata = np.moveaxis(lfpdata,axis,-1)
        spkdata = np.moveaxis(spkdata,time_axis,-1)
        spkdata = np.moveaxis(spkdata,axis,-1)
    data_shape = lfpdata.shape   # Cache data shape after axes shift
    n_timepts,n_obs = data_shape[-2], data_shape[-1] # n_obs = n_trials[*n_tapers]

    # Unwrap all other axes (incl. frequency) -> (n_data_series,n_timepts,n_obs)
    if data_ndim > 2:
        lfpdata = np.reshape(lfpdata, (-1,n_timepts,n_obs))
        spkdata = np.reshape(spkdata, (-1,n_timepts,n_obs))

    # Normalize LFP spectrum/spectrogram so data is all unit-length complex vectors
    lfpdata = lfpdata / np.abs(lfpdata)

    # Set timewins based on given parameters if not set explicitly in args
    if timewins is None:
        timewins = setup_sliding_windows(width,lims,spacing)
    else:
        timewins = np.asarray(timewins)
        width = np.diff(timewins,axis=1).mean()

    # Convert time sampling vector and time windows to int-valued ms,
    #  to avoid floating-point issues in indexing below
    timepts_ms  = np.round(timepts*1000).astype(int)
    timewins_ms = np.round(timewins*1000).astype(int)

    n_data_series = lfpdata.shape[0]
    n_timepts_out = timewins.shape[0]

    vector_mean = np.full((n_data_series,n_timepts_out,1),np.nan,dtype=complex)

    n = np.zeros((n_timepts_out,),dtype=int)

    # Are we computing PLV within temporal windows or at each timepoint
    # (If we don't know smp_rate, assume windowed. Same result, but slower algorithm.)
    do_timewins = (smp_rate is None) or not np.isclose(width, 1/smp_rate)

    # Compute PLV by vector averaging over trials and within given sliding time windows
    if do_timewins:
        for i_win,timewin in enumerate(timewins_ms):
            # Boolean vector flagging all time points within given time window
            tbool = (timepts_ms >= timewin[0]) & (timepts_ms <= timewin[1])

            # Logical AND btwn window and spike train booleans to get spikes in window
            win_spikes = spkdata & tbool[np.newaxis,:,np.newaxis]

            # Count of all spikes within time window across all trials/observations
            n[i_win] = win_spikes.sum()

            # If no spikes in window, can't compute PLV. Skip and leave = nan.
            if n[i_win] == 0: continue

            # Use windowed spike times to index into LFPs and compute complex mean
            # across all spikes (within all trials/observations and window timepoints)
            vector_mean[:,i_win,0] = \
                (lfpdata[np.tile(win_spikes,(n_data_series,1,1))].reshape((n_data_series,n[i_win]))
                                                                 .mean(axis=-1))

            # todo Need to timetest against these alternatives
            # vector_mean[:,i_win] = lfpdata[win_spikes[[0]*n_data_series,:,:]].mean(axis=(-1,-2))
            # vector_mean[:,i_win] = lfpdata[win_spikes[np.zeros((n_data_series,),dtype=int),:,:]].mean(axis=(-1,-2))

    # Compute PLV by vector averaging over trials at each individual timepoint
    else:
        for i_time in range(n_timepts):
            # Count of all spikes within time window across all trials/observations
            n[i_time] = spkdata[0,i_time,:].sum()

            # If no spikes in window, can't compute PLV. Skip and leave = nan.
            if n[i_time] == 0: continue

            # Use spike times to index into LFPs and compute complex mean
            # across all spikes (within all trials/observations)
            vector_mean[:,i_time,0] = lfpdata[:,i_time,spkdata[0,i_time,:]].mean(axis=-1)

    # Reshape axes (incl. frequency) to original data shape
    if data_ndim > 2:
        vector_mean = np.reshape(vector_mean, (*data_shape[:-2],n_timepts_out,1))
    # Move time and trials/observations axes to original locations
    if not ((time_axis == data_ndim-2) and (axis == data_ndim-1)):
        vector_mean = np.moveaxis(vector_mean,-1,axis)
        vector_mean = np.moveaxis(vector_mean,-1,time_axis)

    # HACK If computing PLV as a first step for PPC, temporarily expand dimensionality of n
    # so it will broadcast against PLV (it gets squeezed back down to (n_timewins,) in ppc())
    if _for_ppc:
        slicer = [np.newaxis]*vector_mean.ndim  # Create (PLV.ndim,) list of np.newaxis's
        slicer[time_axis] = slice(None)         # Set <time_axis> element to slice as if set=':'
        if not keepdims: del slicer[axis]       # Remove trial axis if not keeping singleton
        n = n[tuple(slicer)]                    # Expand n to match dimensionality of PLV

    if not keepdims:
        vector_mean = vector_mean.squeeze(axis=axis)

    # Insert singleton axis corresponding to former location of tapers, for dimensional consistency
    elif spec_method == 'multitaper':
        slicer = [slice(None)]*vector_mean.ndim
        slicer = slicer[:time_axis] + [np.newaxis] + slicer[time_axis:]
        vector_mean = vector_mean[tuple(slicer)]

    # Compute absolute value of complex vector mean = mean resultant = PLV
    # and optionally the mean phase angle as well. Also return spike counts.
    if return_phase:
        return np.abs(vector_mean), freqs, timewins.mean(axis=1), n, np.angle(vector_mean)
    else:
        return np.abs(vector_mean), freqs, timewins.mean(axis=1), n


# Alias function with full name
spike_field_phase_locking_value = spike_field_plv
""" Alias of :func:`spike_field_plv`. See there for details. """


def spike_field_ppc(spkdata, lfpdata, axis=0, time_axis=None, taper_axis=None,
                    timepts=None, width=0.5, spacing=None, lims=None, timewins=None,
                    data_type=None, spec_method='wavelet', smp_rate=None,
                    return_phase=False, keepdims=True, **kwargs):
    """
    Compute pairwise phase consistency (PPC) of spike-triggered LFP phase,
    which is bias-corrected (unlike PLV and coherence, which are biased by n)

    PPC is an debiased estimator of PLV^2, and can be expressed (and computed
    efficiently) in terms of PLV and n::

        PPC = (n*PLV^2 - 1) / (n-1)

    Because spiking response are sparse, spike-LFP PPC is typically computed within sliding
    time windows (ie summation across trials AND across within-window timepoints). These can
    be specified either explicitly using `timewins` or implicitly using `width`/`spacing`/`lims`.

    Only parameters differing from :func:`.spike_field_coupling` are described here.

    Parameters
    ----------
    width : scalar, default: 0.5 (500 ms)
        Width (in s) of sliding time windows for computing PPC

    spacing : scalar, default: set = `width` (ie exactly non-overlapping windows)
        Spacing (in s) of sliding time windows for computing PPC

    lims : array-like, shape=(2,), default: (timepts[0],timepts[-1]) (full sampled time of data)
        [Start,end] time limits (in s) for full series of sliding windows

    timewins : ndarray, shape=(n_timewins,2), default: setup_sliding_windows(width,lims,spacing)
        Alternative method for setting sliding time windows; overrides `width`/`spacing`/`lims`.
        Custom time windows to compute PPC within, given as explicit series of window [start,end]'s
        (in s). Can have any arbitrary width/spacing (eg to compute PPC in arbitary time epochs).
        Default generates sliding windows with `width` and `spacing` from `lims[0]` to `lims[1]`

        Special case: Set = (lim[0],lim[1]) to compute PPC *spectrum* over entire data time period.

    **kwargs :
        Any other keyword args passed as-is to spectrogram() function.

    References
    ----------
    - Original concept:   Vinck et al. (2010) https://doi.org/10.1016/j.neuroimage.2010.01.073
    - Relation to PLV:    Kornblith, Buschman, Miller (2015) https://doi.org/10.1093/cercor/bhv182
    """
    extra_args = dict(axis=axis, time_axis=time_axis, taper_axis=taper_axis, timepts=timepts,
                      width=width, spacing=spacing, lims=lims, timewins=timewins,
                      data_type=data_type, spec_method=spec_method, smp_rate=smp_rate,
                      return_phase=return_phase, keepdims=keepdims, _for_ppc=True, **kwargs)

    if return_phase:
        PLV,freqs,timepts,n,phi = spike_field_plv(spkdata, lfpdata, **extra_args)
        return plv_to_ppc(PLV,n), freqs, timepts, n.squeeze(), phi

    else:
        PLV,freqs,timepts,n = spike_field_plv(spkdata, lfpdata, **extra_args)
        return plv_to_ppc(PLV,n), freqs, timepts, n.squeeze()

# Alias function with full name
spike_field_pairwise_phase_consistency = spike_field_ppc
""" Alias of :func:`spike_field_ppc`. See there for details. """


# =============================================================================
# Helper functions
# =============================================================================
def plv_to_ppc(PLV, n):
    """
    Convert PLV to PPC as::

        PPC = (n*PLV^2 - 1)/(n-1)
    """
    return (n*PLV**2 - 1) / (n - 1)



"""
Suite of tests to assess "face validity" of spiking data analysis functions in spikes.py
Usually used to test new or majorly updated functions to ensure they perform as expected.

Includes tests that parametrically estimate spike rate as a function of the simulated data mean,
number of trials, etc. to establish methods produce expected pattern of results.

Plots results and runs assertions that basic expected results are reproduced

Function list
-------------
- test_rate :               Tests of spike rate estimation functions
- rate_test_battery :       Run standard battery of tests of rate functions

- test_rate_stats :         Tests of spike rate statistic estimation functions
- rate_stat_test_battery :  Run standard battery of tests of rate stat functions

- test_isi_stats :          Tests of inter-spike interval statistic estimation functions
- isi_stat_test_battery :   Run standard battery of tests of ISI stat functions

- test_waveform_stats :         Tests of spike waveform statistic estimation functions
- waveform_stat_test_battery :  Run standard battery of tests of spike waveform stat functions
"""
import os
import time
from warnings import warn
from math import sqrt
import numpy as np
import matplotlib.pyplot as plt

from spynal.tests.data_fixtures import simulate_dataset
from spynal.utils import set_random_seed
from spynal.spikes import simulate_spike_trains, simulate_spike_waveforms, times_to_bool, \
                          rate, rate_stats, isi, isi_stats, waveform_stats, \
                          plot_raster, plot_mean_waveforms, plot_waveform_heatmap
from spynal.spectra.multitaper import compute_tapers
from spynal.plots import plot_line_with_error_fill


# =============================================================================
# Tests for rate computation functions
# =============================================================================
def test_rate(method, test='rate', test_values=None, data_type='timestamp', n_trials=1000,
              do_tests=True, do_plots=False, plot_dir=None, seed=1, **kwargs):
    """
    Basic testing for functions estimating spike rate over time.

    Generates synthetic spike train data with given underlying rates,
    estimates rate using given function, and compares estimated to expected.

    For test failures, raises an error or warning (depending on value of `do_tests`).
    Optionally plots summary of test results.

    Parameters
    ----------
    method : str
        Name of rate estimation function to test: 'bin' | 'density'

    test : str, default: 'rate'
        Type of test to run. Options: 'rate' | 'n'

    test_values : array-like, shape=(n_values,), dtype=str
        List of values to test. Interpretation and defaults are test-specific:

        - 'mean' :  Mean spike rate, default: [1,2,5,10,20]
        - 'n' :     Trial numbers, default: [25,50,100,200,400,800]

    data_type : str, default: 'timestamp'
        Type of spiking data to input into rate functions: 'timestamp' | 'bool'

    n_trials : int, default: 1000
        Number of trials to include in simulated data

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    do_plots : bool, default: False
        Set=True to plot test results

    plot_dir : str, default: None (don't save to file)
        Full-path directory to save plots to. Set=None to not save plots.

    seed : int, default: 1 (reproducible random numbers)
        Random generator seed for repeatable results. Set=None for fully random numbers.

    **kwargs :
        All other keyword args passed to rate estimation function

    Returns
    -------
    means : ndarray, shape=(n_rates,)
        Estimated mean rate for each expected rate

    sems : ndarray, shape=(n_rates,)
        SEM of mean rate for each expected rate

    passed : bool
        True if estimated results pass all tests; otherwise False.
    """
    method = method.lower()
    test = test.lower()
    data_type = data_type.lower()

    assert method in ['bin','density'], \
        ValueError("Unsupported value '%s' given for <method>. Should be 'bin' | 'density"
                   % data_type)
    assert data_type in ['timestamp','bool'], \
        ValueError("Unsupported value '%s' given for <data_type>. Should be 'timestamp' | 'bool"
                   % data_type)

    # Set defaults for tested values and set up data generator function depending on <test>
    # Note: Only set random seed once above, don't reset in data generator function calls
    sim_args = dict(offset=10, gain=0.0, data_type='timestamp', n_conds=1, n_trials=n_trials, seed=seed)

    if test in ['rate','mean']:
        test_values = [5,10,20,40] if test_values is None else test_values
        del sim_args['offset']              # Delete preset arg so uses arg to lambda below
        gen_data = lambda rate: simulate_spike_trains(**sim_args, offset=rate)

    elif test in ['n','n_trials']:
        test_values = [25,50,100,200,400,800] if test_values is None else test_values
        del sim_args['n_trials']            # Delete preset arg so uses arg to lambda below
        gen_data = lambda n_trials: simulate_spike_trains(**sim_args, n_trials=n_trials)

    else:
        raise ValueError("Unsupported value '%s' set for <test>" % test)


    if method == 'bin':
        n_timepts = 20
        tbool     = np.ones((n_timepts,),dtype=bool)

    elif method == 'density':
        n_timepts = 1001
        # HACK For spike density method, remove edges, which are influenced by boundary artifacts
        timepts   = np.arange(0,1.001,0.001)
        tbool     = (timepts > 0.1) & (timepts < 0.9)

    else:
        raise ValueError("Unsupported option '%s' given for <method>. \
                         Should be 'bin_rate'|'density'" % method)

    means = np.empty((len(test_values),n_timepts))
    sems = np.empty((len(test_values),n_timepts))

    for i_value,test_value in enumerate(test_values):
        # Generate simulated spike train data
        trains,_ = gen_data(test_value)

        # Convert spike timestamps -> binary 0/1 spike trains (if requested)
        if data_type == 'bool':
            trains,timepts = times_to_bool(trains,lims=[0,1])
            kwargs.update(timepts=timepts)      # Need <timepts> input for bool data

        # Compute spike rate from simulated spike trains -> (n_trials,n_timepts)
        spike_rates,timepts = rate(trains, method=method, lims=[0,1], **kwargs)
        if method == 'bin': timepts = timepts.mean(axis=1)  # bins -> centers

        # Compute mean and SEM of rate time series across trials -> (1,n_timepts)
        n = test_value if test in ['n','n_trials'] else n_trials
        means[i_value,:] = spike_rates.mean(axis=0)
        sems[i_value,:]  = spike_rates.std(axis=0,ddof=0) / sqrt(n)

    # Take average rate across timepoints (excluding ends with edge effect for density method)
    grand_means = means[:,tbool].mean(axis=1)
    grand_sems = sems[:,tbool].mean(axis=1)

    # Optionally plot summary of test results
    if do_plots:
        plt.figure()
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

        # Plot time course of estimated rates for each tested value
        ax = plt.subplot(1,2,1)
        plot_line_with_error_fill(timepts, means, err=sems, ax=ax, color=colors[:len(test_values)])
        for i_value,test_value in enumerate(test_values):
            plt.text(0.99, (0.95-0.05*i_value)*plt.ylim()[1], np.round(test_value,decimals=2),
                     color=colors[i_value], fontweight='bold', horizontalalignment='right')
        plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        plt.xlabel('Time')
        plt.ylabel('Estimated rate')

        # Plot across-time mean rates
        ax = plt.subplot(1,2,2)
        plt.errorbar(test_values, grand_means, grand_sems, marker='o')

        if test in ['rate','mean']:
            ax.set_aspect('equal', 'box')
            plt.plot([0,1.1*test_values[-1]], [0,1.1*test_values[-1]], '-', color='k', linewidth=1)
        plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        plt.xlabel(test)
        plt.ylabel('Estimated rate')
        if plot_dir is not None:
            plt.savefig(os.path.join(plot_dir,'rate-summary-%s-%s-%s.png' % (method,test,data_type)))


    # Determine if test actually produced the expected values
    # 'mean' : Test if estimated rate increases monotonically with actual simulated rate
    if test in ['rate','mean']:
        evals = [((np.diff(grand_means) >= 0).all(),
                    "Estimated rate does not inccrease monotonically with simulated rate")]

    # 'n' : Test if stat is ~ same for all values of n (unbiased by n)
    elif test in ['n','n_trials']:
        evals = [(grand_means.ptp() < grand_sems.max(),
                  "Estimated rate has larger than expected range across n's (likely biased by n)")]

    passed = True
    for cond,message in evals:
        if not cond:    passed = False

        # Raise an error for test fails if do_tests is True
        if do_tests:    assert cond, AssertionError(message)
        # Just issue a warning for test fails if do_tests is False
        elif not cond:  warn(message)

    return means, sems, passed


def rate_test_battery(methods=('bin','density'), tests=('mean','n'),
                      data_types=('timestamp','bool'), do_tests=True, **kwargs):
    """
    Run a battery of given tests on given spike rate computation methods

    Parameters
    ----------
    methods : array-like of str, default: ('bin','density') (all supported methods)
        List of spike rate computation methods to evaluate.

    tests : array-like of str, default: ('mean','n') (all supported tests)
        List of tests to run.

    data_types : array-like of str, default: ('timestamp','bool') (all supported values)
        List of spike data types to evaluate.

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    **kwargs :
        Any other kwargs passed directly to test_rate()
    """
    if isinstance(methods,str): methods = [methods]
    if isinstance(tests,str): tests = [tests]
    if isinstance(data_types,str): data_types = [data_types]

    for method in methods:
        for test in tests:
            for data_type in data_types:
                print("Running %s test on %s %s" % (test,method,data_type))

                t1 = time.time()

                _,_,passed = test_rate(method, test=test, data_type=data_type,
                                       do_tests=do_tests, **kwargs)

                print('%s (test ran in %.1f s)' %
                      ('PASSED' if passed else 'FAILED', time.time()-t1))
                if 'plot_dir' in kwargs: plt.close('all')



# =============================================================================
# Tests for spike rate stats functions
# =============================================================================
def test_rate_stats(stat, test='mean', test_values=None, distribution='poisson', n_trials=1000,
                    n_reps=100, do_tests=True, do_plots=False, plot_dir=None, seed=1, **kwargs):
    """
    Basic testing for functions estimating spike rate statistics

    Generates synthetic spike rate data with given parameters,
    estimates stats using given function, and compares estimated to expected.

    For test failures, raises an error or warning (depending on value of `do_tests`).
    Optionally plots summary of test results.

    Parameters
    ----------
    stat : str
        Name of rate stat to test. Options: 'Fano' | 'CV'

    test : str, default: 'rate'
        Type of test to run. Options: 'rate' | 'spread'

    test_values : array-like, shape=(n_values,), dtype=str
        List of values to test. Interpretation and defaults are test-specific:

        - 'mean' :      Mean spike rate. Default: [1,2,5,10,20]
        - 'spread' :    Gaussian SDs for generating rates. Default: [1,2,5,10,20]
        - 'n' :         Trial numbers. Default: [25,50,100,200,400,800]

    distribution : str, default: 'poisson'
        Name of distribution to simulate data from. Options: 'normal' | 'poisson'

    n_trials : int, default: 1000
        Number of trials to simulate data for

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    do_plots : bool, default: False
        Set=True to plot test results

    plot_dir : str, default: None (don't save to file)
        Full-path directory to save plots to. Set=None to not save plots.

    seed : int, default: 1 (reproducible random numbers)
        Random generator seed for repeatable results. Set=None for fully random numbers.

    **kwargs :
        All other keyword args passed as-is to `simulate_dataset` or `rate_stats` functions,
        as appropriate

    Returns
    -------
    means : ndarray, shape=(n_test_values,)
        Mean of estimated stat values across reps

    sems : ndarray, shape=(n_test_values,)
        Std dev of estimated stat values across reps

    passed : bool
        True if estimated results pass all tests; otherwise False.
    """
    # Note: Set random seed once here, not for every random data generation loop below
    if seed is not None: set_random_seed(seed)

    stat = stat.lower()
    test = test.lower()
    distribution = distribution.lower()
    if test in ['spread','spreads','sd']:
        assert distribution != 'poisson', \
            "Can't run 'spread' test with Poisson data (variance is fixed ~ mean rate)"

    # Set defaults for tested values and set up data generator function depending on <test>
    # Note: Only set random seed once above, don't reset in data generator function calls
    sim_args = dict(gain=5.0, offset=0.0, spreads=5.0, n_conds=1, n=n_trials,
                    distribution=distribution, seed=None)
    # Override defaults with any simulation-related params passed to function
    for arg in sim_args:
        if arg in kwargs: sim_args[arg] = kwargs.pop(arg)    

    if test in ['mean','rate','gain']:
        test_values = [1,2,5,10,20] if test_values is None else test_values
        del sim_args['gain']                    # Delete preset arg so uses arg to lambda below
        gen_data = lambda mean: simulate_dataset(**sim_args,gain=mean)

    elif test in ['spread','spreads','sd']:
        test_values = [1,2,5,10,20] if test_values is None else test_values
        del sim_args['spreads']                 # Delete preset arg so uses arg to lambda below
        gen_data = lambda spread: simulate_dataset(**sim_args,spreads=spread)

    elif test in ['n','n_trials']:
        test_values = [25,50,100,200,400,800] if test_values is None else test_values
        del sim_args['n']                       # Delete preset arg so uses arg to lambda below
        gen_data = lambda n_trials: simulate_dataset(**sim_args,n=n_trials)

    else:
        raise ValueError("Unsupported value '%s' set for <test>" % test)

    stat_values = np.empty((len(test_values),n_reps))
    for i_value,test_value in enumerate(test_values):
        for i_rep in range(n_reps):
            # Generate simulated data with current test value
            data,_ = gen_data(test_value)

            stat_values[i_value,i_rep] = rate_stats(data, stat=stat, axis=0, **kwargs)

    # Compute mean and std dev across different reps of simulation
    stat_sds    = stat_values.std(axis=1,ddof=0)
    stat_means  = stat_values.mean(axis=1)

    if do_plots:
        plt.figure()
        plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        plt.errorbar(test_values, stat_means, stat_sds, marker='o')
        xlabel = 'n' if test == 'bias' else test
        plt.xlabel(xlabel)
        plt.ylabel("%s(rate)" % stat)
        if plot_dir is not None:
            plt.savefig(os.path.join(plot_dir,'stat-summary-%s-%s-%s' % (stat,test,distribution)))

    # Determine if test actually produced the expected values
    # 'mean' : Test if stat decreases monotonically with mean rate for normal data,
    #           remains ~ same for Poisson
    if test == 'mean':
        if distribution == 'normal':
            evals = [((np.diff(stat_means) <= 0).all(),
                      "%s does not decrease monotonically with increase in mean" % stat)]
        elif distribution == 'poisson':
            evals = [(stat_means.ptp() < stat_sds.max(),
                      "%s has larger than expected range for increase in mean of Poisson data"
                      % stat)]

    # 'spread' : Test if stat increases monotonically with increasing distribution spread (rate SD)
    elif test in ['spread','spreads','sd']:
        evals = [((np.diff(stat_means) >= 0).all(),
                  "%s does not increase monotonically with spread increase" % stat)]

    # 'n' : Test if stat is ~ same for all values of n (unbiased by n)
    elif test in ['n','n_trials']:
        evals = [(stat_means.ptp() < stat_sds.max(),
                  "%s has larger than expected range across n's (likely biased by n)" % stat)]

    passed = True
    for cond,message in evals:
        if not cond:    passed = False

        # Raise an error for test fails if do_tests is True
        if do_tests:    assert cond, AssertionError(message)
        # Just issue a warning for test fails if do_tests is False
        elif not cond:  warn(message)

    return stat_means, stat_sds, passed


def rate_stat_test_battery(stats=('fano','cv'), tests=('mean','spread','n'),
                           distributions=('normal','poisson'), do_tests=True, **kwargs):
    """
    Run a battery of given tests on given spike rate statistic computation methods

    Parameters
    ----------
    stats : array-like of str, default: ('fano','cv') (all supported methods)
        List of spike rate stats to evaluate

    tests : array-likeof str, default: ('mean','spread','n') (all supported tests)
        List of tests to run

    distributions : array-like of str, default: ('normal','poisson') (all supported tests)
        List of data distributions to test

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail.

    **kwargs :
        Any other kwargs passed directly to test_rate_stats()
    """
    if isinstance(stats,str): stats = [stats]
    if isinstance(tests,str): tests = [tests]

    for stat in stats:
        for test in tests:
            for distribution in distributions:
                print("Running %s test on %s %s" % (test,stat,distribution))

                # Skip test of distribution spread (SD) for Poisson, bc SD is defined by mean
                if (test.lower() == 'spread') and (distribution.lower() == 'poisson'): continue

                t1 = time.time()

                _,_,passed = test_rate_stats(stat, test=test, distribution=distribution,
                                             do_tests=do_tests, **kwargs)

                print('%s (test ran in %.1f s)' %
                      ('PASSED' if passed else 'FAILED', time.time()-t1))
                if 'plot_dir' in kwargs: plt.close('all')


# =============================================================================
# Tests for ISI stats functions
# =============================================================================
def test_isi_stats(stat, test='mean', test_values=None, n_reps=100,
                   do_tests=True, do_plots=False, plot_dir=None, seed=1, **kwargs):
    """
    Basic testing for functions estimating inter-spike interval statistics

    Generates synthetic spike train data with given parameters,
    estimates ISI stats using given function, and compares estimated to expected.

    For test failures, raises an error or warning (depending on value of `do_tests`).
    Optionally plots summary of test results.

    Parameters
    ----------
    stat : str
        Name of ISI stat to test. Options: 'Fano' | 'CV' | 'CV2 | 'LV' | 'burst_fract'

    test : str, default: 'mean'
        Type of test to run. Only option is 'mean' currently.

    test_values : array-like, shape=(n_values,), dtype=str
        List of values to test. Interpretation and defaults are test-specific:

        - 'mean' : Mean spike rate. Default: [1,2,5,10,20]

    distribution : str, default: 'poisson'
        Name of distribution to simulate data from. Options: 'normal' | 'poisson'

    n_trials : int, default: 1000
        Number of trials to simulate data for

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    do_plots : bool, default: False
        Set=True to plot test results

    plot_dir : str, default: None (don't save to file)
        Full-path directory to save plots to. Set=None to not save plots.

    seed : int, default: 1 (reproducible random numbers)
        Random generator seed for repeatable results. Set=None for fully random numbers.

    **kwargs :
        All other keyword args passed as-is to `simulate_spike_trains` or `isi_stats` functions,
        as appropriate

    Returns
    -------
    means : ndarray, shape=(n_test_values,)
        Mean of estimated stat values across reps

    sems : ndarray, shape=(n_test_values,)
        Std dev of estimated stat values across reps

    passed : bool
        True if estimated results pass all tests; otherwise False.
    """
    # Note: Set random seed once here, not for every random data generation loop below
    if seed is not None: set_random_seed(seed)

    stat = stat.lower()
    test = test.lower()

    # Set defaults for tested values and set up data generator function depending on <test>
    # Set up to run simulation for 10 s to get good estimates, and for n_reps separate trials
    # Note: Only set random seed once above, don't reset in data generator function calls
    sim_args = dict(gain=0.0, offset=5.0, n_conds=1, n_trials=n_reps, time_range=10.0, seed=None)
    # Override defaults with any simulation-related params passed to function
    for arg in sim_args:
        if arg in kwargs: sim_args[arg] = kwargs.pop(arg)

    if test in ['mean','rate','gain']:
        test_values = [1,2,5,10,20] if test_values is None else test_values
        del sim_args['offset']              # Delete preset arg so it uses argument to lambda below
        gen_data = lambda mean: simulate_spike_trains(**sim_args,offset=float(mean))

    else:
        raise ValueError("Unsupported value '%s' set for <test>" % test)

    stat_values = np.empty((len(test_values),n_reps))
    for i_value,test_value in enumerate(test_values):
        # Generate simulated spike timestamp data with current test value
        data,_  = gen_data(test_value)
        ISIs    = isi(data)

        stat_values[i_value,:] = isi_stats(ISIs, stat=stat, axis='each', **kwargs)

    # Compute mean and std dev across different reps of simulation
    stat_sds    = stat_values.std(axis=1,ddof=0)
    stat_means  = stat_values.mean(axis=1)

    if do_plots:
        plt.figure()
        plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        plt.errorbar(test_values, stat_means, stat_sds, marker='o')
        xlabel = 'n' if test == 'bias' else test
        plt.xlabel(xlabel)
        plt.ylabel("%s(ISI)" % stat)
        if plot_dir is not None:
            plt.savefig(os.path.join(plot_dir,'stat-summary-%s-%s' % (stat,test)))

    # Determine if test actually produced the expected values
    # 'mean' : Test if stat remains ~ same for Poisson
    if test == 'mean':
        evals = [(stat_means.ptp() < stat_sds.max(),
                 "%s has larger than expected range for increase in mean of Poisson data" % stat)]

    passed = True
    for cond,message in evals:
        if not cond:    passed = False

        # Raise an error for test fails if do_tests is True
        if do_tests:    assert cond, AssertionError(message)
        # Just issue a warning for test fails if do_tests is False
        elif not cond:  warn(message)

    return stat_means, stat_sds, passed


def isi_stat_test_battery(stats=('Fano','CV','CV2','LV','burst_fract'), tests=('mean'),
                          do_tests=True, **kwargs):
    """
    Run a battery of given tests on given inter-spike interval statistic computation methods

    Parameters
    ----------
    stats : array-like of str, default: ('Fano','CV','CV2','LV','burst_fract') (all supported)
        List of ISI stats to evaluate

    tests : array-like of str, default: ('mean') (all supported tests)
        List of tests to run

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail.

    **kwargs :
        Any other kwargs passed directly to test_isi_stats()
    """
    if isinstance(stats,str): stats = [stats]
    if isinstance(tests,str): tests = [tests]

    for stat in stats:
        for test in tests:
            print("Running %s test on %s" % (test,stat))
            do_tests_ = False if (stat == 'burst_fract') and (test == 'mean') else do_tests

            t1 = time.time()

            _,_,passed = test_isi_stats(stat, test=test, do_tests=do_tests_, **kwargs)

            print('%s (test ran in %.1f s)' % ('PASSED' if passed else 'FAILED', time.time()-t1))
            if 'plot_dir' in kwargs: plt.close('all')


# =============================================================================
# Tests for spike waveform stats functions
# =============================================================================
def test_waveform_stats(stat, test='width', test_values=None, n_spikes=100,
                        n_reps=100, do_tests=True, do_plots=False, plot_dir=None,
                        seed=1, **kwargs):
    """
    Basic testing for functions estimating spike waveform statistics

    Generates synthetic spike waveform data with given parameters,
    estimates stats using given function, and compares estimated to expected.

    For test failures, raises an error or warning (depending on value of `do_tests`).
    Optionally plots summary of test results.

    Parameters
    ----------
    stat : str
        Name of rate stat to test. Options: 'width' | 'trough_width' | 'peak_width' | 'trough_amp'

    test : str, default: 'width'
        Type of test to run. Options: 'width' | 'trough_width' | 'repol' | 'amp_ratio'

    test_values : array-like, shape=(n_values,), dtype=str
        List of values to test. Interpretation and defaults are test-specific:

        - 'width' :         Trough-to-peak spike width (s). Default: [0.75,1,1.25]*1e-3
        - 'trough_width' :  Full width (2*SD; s) of waveform trough. Default: [0.12,0.25,0.37]*1e-3
        - 'peak_width' :    Full width (2*SD; s) of waveform trough. Default: [0.37,0.5,0.62]*1e-3
        - 'trough_amp' :    Amplitude of waveform trough (mV). Default: [0.2,0.4,0.8]

    n_spikes : int, default: 100
        Number of spike waveforms to simulate

    n_reps : int, default: 100
        Number of repetitions to run simulation & test for

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    do_plots : bool, default: False
        Set=True to plot test results

    plot_dir : str, default: None (don't save to file)
        Full-path directory to save plots to. Set=None to not save plots.

    seed : int, default: 1 (reproducible random numbers)
        Random generator seed for repeatable results. Set=None for fully random numbers.

    **kwargs :
        All other keyword args passed as-is to `simulate_spike_waveforms` or `waveform_stats`
        functions, as appropriate

    Returns
    -------
    means : ndarray, shape=(n_test_values,)
        Mean of estimated stat values across reps

    sems : ndarray, shape=(n_test_values,)
        Std dev of estimated stat values across reps

    passed : bool
        True if estimated results pass all tests; otherwise False.
    """
    # Note: Set random seed once here, not for every random data generation loop below
    if seed is not None: set_random_seed(seed)

    stat = stat.lower()
    test = test.lower()

    # Set defaults for tested values and set up data generator function depending on <test>
    # Note: Only set random seed once above, don't reset in data generator function calls
    # todo Should we move some/all of these into function arguments, instead of hard-coding?
    sim_args = dict(trough_time=0.3e-3, peak_time=0.75e-3, trough_amp=0.4, peak_amp=0.25,
                    trough_width=0.2e-3, peak_width=0.4e-3, time_range=(-0.2e-3,1.4e-3),
                    smp_rate=30e3, noise=0.01, n_spikes=100, seed=None)
    # Override defaults with any simulation-related params passed to function
    for arg in sim_args:
        if arg in kwargs: sim_args[arg] = kwargs.pop(arg)

    if test in ['width','t2p','trough_to_peak_width']:
        test_values = np.asarray([0.35,0.45,0.55])*1e-3 if test_values is None else test_values
        del sim_args['peak_time']               # Delete preset arg so uses arg to lambda below
        gen_data = lambda width: simulate_spike_waveforms(**sim_args,peak_time=width+sim_args['trough_time'])

    elif test in ['trough_width','trough']:
        test_values = np.asarray([0.15,0.2,0.25])*1e-3 if test_values is None else test_values
        del sim_args['trough_width']            # Delete preset arg so uses arg to lambda below
        gen_data = lambda trough_width: simulate_spike_waveforms(**sim_args,trough_width=trough_width)

    elif test in ['peak_width','peak']:
        test_values = np.asarray([0.3,0.4,0.5])*1e-3 if test_values is None else test_values
        del sim_args['peak_width']              # Delete preset arg so uses arg to lambda below
        gen_data = lambda peak_width: simulate_spike_waveforms(**sim_args,peak_width=peak_width)

    elif test in ['trough_amp','amp']:
        test_values = [0.2,0.4,0.8] if test_values is None else test_values
        del sim_args['trough_amp']              # Delete preset arg so uses arg to lambda below
        gen_data = lambda trough_amp: simulate_spike_waveforms(**sim_args,trough_amp=trough_amp)

    else:
        raise ValueError("Unsupported value '%s' set for <test>" % test)

    if stat not in ['trough_peak_amp_ratio','amp_ratio']: kwargs.update(smp_rate=30e3)

    _,timepts = gen_data(test_values[0])
    waveform_means = np.empty((len(timepts), len(test_values), n_reps))
    stat_values = np.empty((len(test_values), n_reps))

    for i_value,test_value in enumerate(test_values):
        for i_rep in range(n_reps):
            # Generate simulated data with current test value and average across all sim'd spikes
            data,_ = gen_data(test_value)
            mean_waveform = data.mean(axis=1)
            waveform_means[:,i_value,i_rep] = mean_waveform

            stat_values[i_value,i_rep] = waveform_stats(mean_waveform, stat=stat, axis=0, **kwargs)

    # Compute mean and std dev across different reps of simulation
    stat_sds    = stat_values.std(axis=1,ddof=0)
    stat_means  = stat_values.mean(axis=1)

    if do_plots:
        # Convert time values from s -> us for plotting clarity
        test_mult = 1 if 'amp' in test else 1e6
        stat_mult = 1 if 'amp' in stat else 1e6

        plt.figure()
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

        ax = plt.subplot(1,2,1)
        plot_line_with_error_fill(timepts*1e6, waveform_means.mean(axis=-1).T,
                                  err=waveform_means.std(axis=-1).T, ax=ax,
                                  color=colors[:len(test_values)])
        for i_value,test_value in enumerate(test_values):
            plt.text(0.99, (0.95-0.05*i_value), np.round(test_value*test_mult,decimals=1),
                     color=colors[i_value], fontweight='bold', horizontalalignment='right',
                     transform=ax.transAxes)
        plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        plt.xlabel('Time (us)')
        plt.ylabel('Voltage (mV)')

        plt.subplot(1,2,2)
        plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        plt.errorbar(test_values*test_mult, stat_means*stat_mult, stat_sds*stat_mult, marker='o')
        xlabel = 'n' if test == 'bias' else test
        plt.xlabel(xlabel)
        plt.ylabel(stat)
        if plot_dir is not None:
            plt.savefig(os.path.join(plot_dir,'stat-summary-%s-%s' % (stat,test)))


    # Determine if test actually produced the expected values
    # 'width' : Test if t2p stat increases monotonically with width, other stats don't
    if test in ['width','t2p','trough_to_peak_width']:
        if stat in ['width','t2p','trough_to_peak_width']:
            evals = [((np.diff(stat_means) > 0).all(),
                      "%s does not increase monotonically with increase in spike width" % stat)]
        else:
            evals = [(stat_means.ptp() < stat_sds.max(),
                      "%s has larger than expected range for changes in spike width" % stat)]

    # 'trough_width' : Test if trough_width stat increases monotonically, other stats don't
    elif test in ['trough_width','trough']:
        if stat in ['trough_width','trough']:
            evals = [((np.diff(stat_means) > 0).all(),
                      "%s does not increase monotonically with increase in trough width" % stat)]
        else:
            evals = [(stat_means.ptp() < stat_sds.max(),
                      "%s has larger than expected range for changes in trough width" % stat)]

    # 'peak_width' : Test if repolarization stat increases monotonically, other stats don't
    elif test in ['peak_width','peak']:
        if stat in ['repol','repolarization','repolarization_time']:
            evals = [((np.diff(stat_means) > 0).all(),
                      "%s does not increase monotonically with increase in peak width" % stat)]
        else:
            evals = [(stat_means.ptp() < stat_sds.max(),
                      "%s has larger than expected range for changes in peak width" % stat)]

    # 'trough_amp' : Test if amp_ratio stat increases monotonically, other stats don't
    elif test in ['trough_amp','amp']:
        if stat in ['trough_peak_amp_ratio','amp_ratio']:
            evals = [((np.diff(stat_means) > 0).all(),
                      "%s does not increase monotonically with increase in trough amp" % stat)]
        else:
            evals = [(stat_means.ptp() < stat_sds.max(),
                      "%s has larger than expected range for changes in trough amp" % stat)]

    passed = True
    for cond,message in evals:
        if not cond:    passed = False

        # Raise an error for test fails if do_tests is True
        if do_tests:    assert cond, AssertionError(message)
        # Just issue a warning for test fails if do_tests is False
        elif not cond:  warn(message)

    return stat_means, stat_sds, passed


def waveform_stat_test_battery(stats=('width','trough_width','repol','amp_ratio'),
                               tests=('width','trough_width','peak_width','trough_amp'),
                               do_tests=True, **kwargs):
    """
    Run a battery of given tests on given spike waveform statistic computation methods

    Parameters
    ----------
    stats : array-like of str, default: ('width','trough_width','repol','amp_ratio')
        List of spike waveform stats to evaluate

    tests : array-likeof str, default: ('width','trough_width','peak_width','trough_amp')
        List of tests to run

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail.

    **kwargs :
        Any other kwargs passed directly to test_waveform_stats()
    """
    if isinstance(stats,str): stats = [stats]
    if isinstance(tests,str): tests = [tests]

    for stat in stats:
        for test in tests:
            print("Running %s test on %s" % (test,stat))

            t1 = time.time()

            _,_,passed = test_waveform_stats(stat, test=test, do_tests=do_tests, **kwargs)

            print('%s (test ran in %.1f s)' %
                    ('PASSED' if passed else 'FAILED', time.time()-t1))
            if 'plot_dir' in kwargs: plt.close('all')


# =============================================================================
# Tests for plotting functions
# =============================================================================
def test_plot_raster(plot_dir=None):
    """
    Basic testing for plotting function plot_raster()

    Parameters
    ----------
    plot_dir : str, default: None (don't save plots)
        Full-path directory to save plots to. Set=None to not save plots.

    Actions
    -------
        Creates a plot and optionally saves it to PNG file
    """
    data, _ = simulate_spike_trains(n_trials=100, offset=20, n_conds=1, refractory=1e-3, seed=1)

    for data_type in ['timestamp','bool']:
        if data_type == 'bool':
            data, timepts = times_to_bool(data, lims=[0,1])

        for graphics in ['vector','bitmap']:
            print(data_type,graphics)
            t1 = time.time()
            extra_args = dict(graphics=graphics)
            if data_type == 'bool': extra_args.update(timepts=timepts)
            else:                   extra_args.update(lims=(0,1))
            plot_str = '-' + data_type + '-' + graphics

            # Basic test plot
            plt.figure()
            ax = plot_raster(data, **extra_args)
            plt.title('Basic test raster plot (%s data, %s graphics)' % (data_type,graphics), fontsize='small')
            plt.show()
            if plot_dir is not None: plt.savefig(os.path.join(plot_dir,'plot_raster'+plot_str+'.png'))

            # Plot rasters in different color
            ax = plot_raster(data, **extra_args, color=[0.5,0.0,0.0])
            plt.title('Modified color raster plot (%s data, %s graphics)' % (data_type,graphics), fontsize='small')
            plt.show()
            if plot_dir is not None: plt.savefig(os.path.join(plot_dir,'plot_raster'+plot_str+'-color.png'))

            # Plot rasters with different height of plotted spikes
            ax = plot_raster(data, **extra_args, height=0.5)
            plt.title('Modified height raster plot (%s data, %s graphics)' % (data_type,graphics), fontsize='small')
            plt.show()
            if plot_dir is not None: plt.savefig(os.path.join(plot_dir,'plot_raster'+plot_str+'-height.png'))

            # Plot rasters with overlaid event markers
            ax = plot_raster(data, **extra_args, events=[(100e-3,300e-3), 500e-3, (880e-3,900e-3,920e-3)])
            plt.title('Raster with events plot (%s data, %s graphics)' % (data_type,graphics), fontsize='small')
            plt.show()
            if plot_dir is not None: plt.savefig(os.path.join(plot_dir,'plot_raster'+plot_str+'-events.png'))

            print("Finished all tests in %.3f s" % (time.time()-t1))


def test_plot_mean_waveforms(plot_dir=None):
    """
    Basic testing for plotting function plot_mean_waveforms()

    Parameters
    ----------
    plot_dir : str, default: None (don't save plots)
        Full-path directory to save plots to. Set=None to not save plots.

    Actions
    -------
        Creates a plot and optionally saves it to PNG file
    """
    n_units = 3
    n_spikes = 100
    n_timepts = 1000

    # Use dpss tapers as data to plot, since they are all very distinct looking
    means = compute_tapers(n_timepts, time_width=1.0, freq_width=4, n_tapers=n_units)
    waveforms = np.empty((n_units,), dtype=object)
    for unit in range(n_units):
        waveforms[unit] = means[:,[unit]] + np.random.standard_normal((n_timepts,n_spikes))

    # Basic test plot
    plt.figure()
    plot_mean_waveforms(waveforms, plot_sd=True)
    plt.title('Basic test plot')
    plt.show()
    if plot_dir is not None: plt.savefig(os.path.join(plot_dir,'plot_mean_waveforms.png'))

    # Plot w/o SDs (means only)
    plt.figure()
    plot_mean_waveforms(waveforms, plot_sd=False)
    plt.title('Means-only plot')
    plt.show()
    if plot_dir is not None: plt.savefig(os.path.join(plot_dir,'plot_mean_waveforms-noSD.png'))


def test_plot_waveform_heatmap(plot_dir=None):
    """
    Basic testing for plotting function plot_waveform_heatmap()

    Parameters
    ----------
    plot_dir : str, default: None (don't save plots)
        Full-path directory to save plots to. Set=None to not save plots.

    Actions
    -------
        Creates a plot and optionally saves it to PNG file
    """
    n_units = 1
    n_spikes = 100
    n_timepts = 1000

    # Use dpss tapers as data to plot, since they are all very distinct looking
    means = compute_tapers(n_timepts, time_width=1.0, freq_width=4, n_tapers=n_units)
    waveforms = np.empty((n_units,), dtype=object)
    for unit in range(n_units):
        waveforms[unit] = means[:,[unit]] + np.random.standard_normal((n_timepts,n_spikes))

    # Basic test plot
    plt.figure()
    plot_waveform_heatmap(waveforms)
    plt.title('Basic test plot')
    if plot_dir is not None: plt.savefig(os.path.join(plot_dir,'plot_waveform_heatmap.png'))

    # Basic test plot
    plt.figure()
    plot_waveform_heatmap(waveforms, n_ybins=50)
    plt.title('Fine-grained bins')
    if plot_dir is not None: plt.savefig(os.path.join(plot_dir,'plot_waveform_heatmap-50bins.png'))



# -*- coding: utf-8 -*-
""" Utility functions for LFP/EEG/continuous data and spectral analysis """
from math import pi, ceil, log2
import numpy as np

from scipy.stats import norm

from spynal.utils import axis_index_slices, set_random_seed, interp1


def next_power_of_2(n):
    """ Find next power of 2 (smallest power of 2 greater than n) """
    # todo  Think about switching this to use scipy.fftpack.next_fast_len
    return 1 if n == 0 else 2**ceil(log2(n))


def get_freq_sampling(smp_rate, n_fft, freq_range=None, two_sided=False):
    """
    Return frequency sampling vector (axis) for a given FFT-based computation

    Parameters
    ----------
    smp_rate : scalar
        Data sampling rate (Hz)

    n_fft : scalar
        Number of samples (timepoints) in FFT output

    freq_range : array-like, shape=(2,) or scalar, default: all frequencies from FFT
        Range of frequencies to retain in output, either given as an explicit [low,high]
        range or just a scalar giving the highest frequency to return.


    two_sided : bool, default: False
        If True, return freqs for two-sided spectrum, including both positive and
        negative frequencies (which have same amplitude for all real signals).
        If False, only return positive frequencies, in range (0,smp_rate/2).

    Returns
    -------
    freqs : ndarray, shape=(n_freqs,)
        Frequency sampling vector (in Hz)

    freq_bool : ndarray, shape=(n_fft,), dtype=bool
        Boolean vector flagging frequencies in full FFT output to retain, given desired freq_range
    """
    freqs   = np.fft.fftfreq(n_fft,d=1/smp_rate) # All possible frequencies

    # If no range requested, keep all frequencies
    if freq_range is None:
        # Include both positive and negative frequencies
        if two_sided:
            freq_bool = np.ones((n_fft,),dtype=bool)
        # Limit to positive frequencies
        else:
            if n_fft%2 == 0: n = (n_fft/2 + 1, n_fft/2 - 1)
            else:            n = ((n_fft-1)/2, (n_fft-1)/2 + 1)
            freq_bool = np.concatenate((np.ones((int(n[0]),),dtype=bool),
                                        np.zeros((int(n[1]),),dtype=bool)))

    # Limit frequencies to requested range
    else:
        # Only keep frequencies < max freq, or w/in given range
        if len(freq_range) == 1:
            freq_bool = np.abs(freqs) <= freq_range
        elif len(freq_range) == 2:
            freq_bool = (np.abs(freqs) >= freq_range[0]) & \
                        (np.abs(freqs) <= freq_range[1])
        else:
            raise ValueError("freq_range must be given as 2-length vector = [min,max]" \
                             "or scalar max frequency")

        # Limit to positive frequencies. Special case to also get f = (-)smp_rate/2
        if not two_sided:
            freq_bool = freq_bool & ((freqs >= 0) | np.isclose(freqs,-smp_rate/2))

    # Extract only desired freqs from sampling vector
    freqs = freqs[freq_bool]

    # Again, special case to deal with (-)smp_rate/2
    if not two_sided: freqs = np.abs(freqs)

    return freqs,freq_bool


def complex_to_spec_type(data, spec_type):
    """
    Converts complex spectral data to given spectral signal type

    Parameters
    ----------
    data : ndarray, shape=Any, dtype=complex
        Complex spectral (or time-frequency) data. Arbitrary shape.

    spec_type : {'power','phase','magnitude','real','imag'}
        Type of spectral signal to return:

        - 'power'     Spectral power of data
        - 'phase'     Phase of complex spectral data (in radians)
        - 'magnitude' Magnitude (square root of power) of complex data = signal envelope
        - 'real'      Real part of complex data
        - 'imag'      Imaginary part of complex data

    Returns
    -------
    data : ndarray, shape=Any, dtype=complex
        Computed spectral signal. Same shape as input.
    """
    if spec_type == 'complex':      return data
    elif spec_type == 'power':      return power(data)
    elif spec_type == 'phase':      return phase(data)
    elif spec_type == 'magnitude':  return magnitude(data)
    elif spec_type == 'real':       return data.real
    elif spec_type == 'imag':       return np.imag(data)
    else:
        raise ValueError("%s is an unsupported option for spec_type" % spec_type)


def power(data):
    """ Compute power from complex spectral data  """
    return (data*data.conj()).real  # Note: .real fixes small float errors


def magnitude(data):
    """ Compute magnitude (square root of power) from complex spectral data  """
    return np.abs(data)


def phase(data):
    """ Compute phase of complex spectral data  """
    return np.angle(data)


def real(data):
    """ Return real part of complex spectral data  """
    return data.real


def imag(data):
    """ Return imaginary part of complex spectral data  """
    return np.imag(data)


def one_sided_to_two_sided(data, freqs, smp_rate, axis=0):
    """
    Convert a one-sided Fourier/wavelet transform output to the two-sided equivalent.

    Assumes conjugate symmetry across positive and negative frequencies (as is the case
    only when the original raw signals were real).

    Also extrapolates values for f=0, as is necessary for wavelet transforms.

    Parameters
    ----------
    data : ndarray, shape=(...,n_freqs,...), dtype=complex
        Complex (1-sided) frequency-transformed data. Any arbitary shape.

    freqs : array-like, shape=(n_freqs,)
        Frequency sampling in `data`

    smp_rate : scalar
        Data sampling rate (Hz)

    axis : int, default: 0 (1st axis)
        Data axis corresponding to frequency

    Returns
    -------
    data : ndarray, shape=(...,2*n_freqs+1,...), dtype=complex
        2-sided equivalent of input `data`

    freqs : ndarray, shape=(2*n_freqs+1,)
        List of (positive and negative) freqs in 2-sided output `data`
    """
    assert np.isclose(freqs[-1],smp_rate/2), \
        "Need to have sampling up to 1/2 sampling rate (Nyquist freq=%d Hz)" % (smp_rate/2)

    # If f=0 is not in data, numerically extrapolate values for it
    if not np.isclose(freqs,0).any():
        f0 = interp1(freqs, data, 0, axis=axis, kind='cubic', fill_value='extrapolate')
        f0 = np.expand_dims(f0,axis)
        data = np.concatenate((f0,data),axis=axis)
        freqs = np.concatenate(([0],freqs))

    # Convert values at Nyquist freq to complex conjugate at negative frequency
    slices = axis_index_slices(axis,-1,data.ndim)
    data[slices] = data[slices].conj()
    freqs[-1] *= -1

    # Replicate values for all freqs (s.t. 0 < f < nyquist)
    # as complex conjugates at negative frequencies
    idxs    = slice(-2,1,-1)
    slices  = axis_index_slices(axis,idxs,data.ndim)
    data    = np.concatenate((data, data[slices].conj()), axis=axis)
    freqs   = np.concatenate((freqs, -freqs[idxs]))

    return data, freqs


def simulate_oscillation(frequency, amplitude=5.0, phase=0, noise=1.0, n_trials=1000,
                         freq_sd=0, amp_sd=0, phase_sd=0,
                         smp_rate=1000, time_range=1.0, burst_rate=0, burst_width=4, seed=None):
    """
    Generate synthetic data with oscillation at given parameters.

    Generate multiple trials with constant oscillatory signal + random additive Gaussian noise.

    Parameters
    ----------
    frequency : scalar
        Frequency to simulate oscillation at (Hz)

    amplitude : scalar, default: 5.0
        Amplitude of simulated oscillation (a.u.)

    phase : scalar, default: 0
        Phase of oscillation (rad)

    noise : scalar, default: 1.0
        Amplitude of additive Gaussian noise (a.u)

    n_trials : int, default: 1000
        Number of trials/observations to simulate

    freq_sd,amp_sd,phase_sd : scalar, Default: 0 (no inter-trial variation)
        Inter-trial variation in frequency/amplitude/phase, given as Gaussian SD
        (same units as base parameters, which are used as Gaussian mean)

    smp_rate : int, default: 1000
        Sampling rate for simulated data (Hz)

    time_range : scalar, default: 1 s
        Full time range to simulate oscillation over (s)

    burst_rate : scalar, default: 0 (not bursty)
        Oscillatory burst rate (bursts/trial). Set=0 to simulate constant, non-bursty oscillation.

    burst_width : scalar, default: 4
        Half-width of oscillatory bursts (Gaussian SD, in cycles)

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    Returns
    -------
    data : ndarray, shape=(n_timepts,n_trials)
        Simulated oscillation-in-noise data
    """
    if seed is not None: set_random_seed(seed)

    def _randn(*args):
        """
        Generate unit normal random variables in a way that reproducibly matches output
        of Matlab with same seed != 0. (np.random.randn() does not work here for unknown reasons)
        stackoverflow.com/questions/3722138/is-it-possible-to-reproduce-randn-of-matlab-with-numpy?noredirect=1&lq=1
        """
        return norm.ppf(np.random.rand(*args))

    # Set per-trial frequency, amplitude, phase from base parameter + any spread
    freq    = frequency if freq_sd == 0 else frequency + freq_sd*_randn(1,n_trials)
    amp     = amplitude if amp_sd == 0 else amplitude + amp_sd*_randn(1,n_trials)
    phi     = phase if phase_sd == 0 else phase + phase_sd*_randn(1,n_trials)

    # Simulate oscillatory bursts if burst_rate is set != 0
    bursty = burst_rate > 0
    # Convert burst width from cycles to s
    burst_sd = burst_width/freq

    # Set time sampling vector (in s)
    n_timepts = round(time_range * smp_rate)
    t = np.arange(n_timepts) / smp_rate

    # Generate oscillatory signal = sinusoid wave at given amplitude(s),frequency(s),phase(s)
    if np.isscalar(amp) and np.isscalar(freq) and np.isscalar(phi):
        data = np.tile((amp * np.cos(2*pi*freq*t + phi))[:,np.newaxis], (1,n_trials))
    else:
        data = amp * np.cos(2*pi*freq*t[:,np.newaxis] + phi)

    # Make oscillations bursty, if requested
    if bursty:
        # Function to generate unit-height Gaussian function with given mean,SD
        def _gaussian(mu,sd,t):
            z = (t - mu)/sd
            return np.exp(-0.5*(z**2))

        # Use burst rate to determine which trials will have bursts ~ Bernoulli(p=rate)
        burst_trials = np.random.binomial(1,burst_rate, size=(n_trials,)).astype(bool)

        # Generate random burst times within full time range of data
        burst_times = np.empty((n_trials,))
        burst_times[burst_trials] = t[0] + (t[-1]-t[0]) * np.random.rand(burst_trials.sum())

        # Weight current trial by random Gaussian envelope if it has a burst
        # otherwise, remove signal from non-burst trials
        for trial in range(n_trials):
            if burst_trials[trial]:
                data[:,trial] *= _gaussian(burst_times[trial],burst_sd,t)
            else:
                data[:,trial] *= 0

    # Generate additive Gaussian noise of given amplitude
    if noise != 0:  data += noise * _randn(n_timepts,n_trials)

    return data
